import abc
from abc import abstractmethod
from collections.abc import Iterable
import itertools
import math
from typing import Literal, TYPE_CHECKING

try:
    from typing import override  # py>=3.12
except ImportError:  # pragma: no cover
    from typing_extensions import override

from localuf import constants
from localuf.noise import CodeCapacity
from localuf.type_aliases import Edge, Node
from localuf._pairs import LogicalCounter, Pairs
from localuf._determinants import Determinant, SpaceDeterminant, SpaceTimeDeterminant


if TYPE_CHECKING:
    from localuf._base_classes import Code, Decoder
    from localuf.decoders.snowflake.main import Snowflake


class Scheme(abc.ABC):
    """Abstract base class for decoding scheme of a CSS code."""

    def __init__(self, code: 'Code'):
        """Input: ``code`` the CSS code."""
        self._CODE = code
        """The CSS code."""
        self._DETERMINANT: Determinant
        """Object to determine whether a node is a boundary."""

    @property
    @abc.abstractmethod
    def WINDOW_HEIGHT(self) -> int:
        """Total height of decoding graph G or sliding window W."""

    @abc.abstractmethod
    def get_logical_error(self, leftover: set[Edge]) -> int:
        """See ``Code.get_logical_error``."""

    def is_boundary(self, v: Node):
        """See ``Code.is_boundary``."""
        return self._DETERMINANT.is_boundary(v)

    @abc.abstractmethod
    def run(
        self,
        decoder: 'Decoder',
        noise_level: float,
        n: int,
        **kwargs,
    ) -> tuple[int, int | float]:
        """Simulate multiple decoding cycles at a given noise level.
        
        
        :param decoder: The decoder.
        :param noise_level: The noise level.
        :param n: Depends on the decoding scheme.
            If the scheme is 'batch',
            then ``n`` is decoding cycle count.
            If the scheme is 'global batch',
            then ``n`` is slenderness := (layer count / code distance),
            where layer count := measurement round count + 1.
            If the scheme is 'forward',
            then ``n`` is the number of decoding cycles in the steady state.
            If the scheme is 'frugal',
            then ``n`` is decoding cycle count in the steady state, divided by the code distance.
            For Snowflake, the there is 1 decoding cycle per stabiliser measurement round.
        
        :return logical_error_count: Logical error count.
        :return slenderness: Decoding cycle count
            if noise is code capacity; else, (total layer count / code distance).
        """

    @abc.abstractmethod
    def sim_cycles_given_weight(
            self,
            decoder: 'Decoder',
            weight: int | tuple[int, ...],
            n: int,
    ) -> tuple[int, int]:
        """Simulate ``n`` decoding cycles given ``weight``.
        
        
        :param decoder: the decoder.
        :param weight: the weight of the error.
        :param n: decoding cycle count.
        
        :return failure_count: Failure count.
        :return n: decoding cycle count.
        """


class Batch(Scheme):
    """Batch decoding scheme.
    
    Extends ``Scheme``.
    """

    @staticmethod
    def __str__() -> str:
        return 'batch'

    def __init__(self, code: 'Code', window_height: int):
        """Additional inputs: ``window_height`` the height of the batch."""
        super().__init__(code)
        self._WINDOW_HEIGHT = window_height
        self._DETERMINANT = SpaceDeterminant(code.D, code.LONG_AXIS)

    @property
    def WINDOW_HEIGHT(self): return self._WINDOW_HEIGHT

    @override
    def run(self, decoder: 'Decoder', noise_level: float, n: int):
        # this assumes logical error count per batch << 1
        m = sum(self._sim_cycle_given_p(decoder, noise_level) for _ in itertools.repeat(None, n))
        return (m, n if isinstance(self._CODE.NOISE, CodeCapacity)
            else self.WINDOW_HEIGHT * n / self._CODE.D)

    @override
    def get_logical_error(self, leftover):
        """Return logical error count parity in ``leftover``."""
        flip_count: int = 0
        for u, _ in leftover:
            flip_count += (u[self._CODE.LONG_AXIS] == -1)
        return flip_count % 2

    def _sim_cycle_given_p(self, decoder: 'Decoder', noise_level: float) -> int:
        """Simulate a decoding cycle at a given noise level.
        
        
        :param decoder: the decoder.
        :param noise_level: noise level.
        
        :return logical_error: ``0`` if success else ``1``.
        """
        error = self._CODE.make_error(noise_level)
        return self._sim_cycle_given_error(decoder, error)

    def _sim_cycle_given_error(self, decoder: 'Decoder', error: set[Edge]):
        """Simulate a decoding cycle given ``error``.
        
        
        :param error: the set of bitflipped edges.
        
        :return logical_error: ``0`` if success else ``1``.
        """
        syndrome = self._CODE.get_syndrome(error)
        decoder.reset()
        decoder.decode(syndrome)
        leftover = error ^ decoder.correction
        return self.get_logical_error(leftover)

    @override
    def sim_cycles_given_weight(self, decoder, weight, n):
        m = 0
        for _ in itertools.repeat(None, n):
            error = self._CODE.NOISE.force_error(weight)
            m += self._sim_cycle_given_error(decoder, error)
        return m, n


class Global(Batch):
    """Global batch decoding scheme.
    
    Extends ``Batch``.
    """

    @staticmethod
    def __str__() -> str:
        return 'global batch'

    def __init__(self, code: 'Code', window_height: int):
        self.pairs = Pairs()
        """Set of node pairs defining free anyon strings, used to count logical error strings."""
        super().__init__(code, window_height)

    def reset(self):
        """Factory reset."""
        self.pairs.reset()

    @override
    def run(self, decoder: 'Decoder', noise_level: float, n: int):
        # `slenderness = n` assumes the following:
        assert self.WINDOW_HEIGHT == self._CODE.D * n
        self.reset()
        m = self._sim_cycle_given_p(decoder, noise_level)
        return m, n

    @override
    def get_logical_error(self, leftover: set[Edge]):
        """Count logical errors in ``leftover``."""
        for e in leftover:
            self.pairs.load(e)
        error_count: int = 0
        for u, v in self.pairs.as_set:
            pair_separation = abs(u[self._CODE.LONG_AXIS] - v[self._CODE.LONG_AXIS])
            if pair_separation == self._CODE.D:
                error_count += 1
        return error_count


class _Streaming(Scheme):
    """Abstract base class for stream decoding schemes.
    
    Extends ``Scheme``.
    """

    def __init__(
            self,
            code: 'Code',
            commit_height: int,
            buffer_height: int,
            commit_edges: tuple[Edge, ...],
    ):
        """
        :param commit_height: The height of the commit region.
        :param buffer_height: The height of the buffer region.
        :param commit_edges: The edges in the commit region.
        """
        super().__init__(code)
        self._COMMIT_HEIGHT = commit_height
        """The height of the commit region."""
        self._BUFFER_HEIGHT = buffer_height
        """The height of the buffer region."""
        self._COMMIT_EDGES = commit_edges
        """The edges of the commit region."""
        self._BUFFER_EDGES = tuple(set(code.EDGES) - set(commit_edges))
        """The edges of the buffer region."""
        self._DETERMINANT = SpaceTimeDeterminant(
            d=code.D,
            long_axis=code.LONG_AXIS,
            time_axis=code.TIME_AXIS,
            window_height=self.WINDOW_HEIGHT,
        )
        self.pairs = Pairs()
        """Set of node pairs defining free anyon strings, used to count logical error strings."""
        self._LOGICAL_COUNTER = LogicalCounter(
            d=code.D,
            commit_height=commit_height,
            long_axis=code.LONG_AXIS,
            time_axis=code.TIME_AXIS,
        )
        """Object for ``get_logical_error``."""
        self.step_counts = []
        """A list in which each entry is the decoder timestep count of ``d`` decoding cycles.
        Populated when ``run`` is called.
        """

    @property
    def WINDOW_HEIGHT(self): return self._COMMIT_HEIGHT + self._BUFFER_HEIGHT

    @abstractmethod
    def reset(self):
        """Factory reset."""
        self.pairs.reset()
        self.step_counts.clear()

    @override
    def get_logical_error(self):
        """Count logical errors completed in current commit.
        
        
        :return logical_error_count: The number of logical errors,
            i.e. paths between opposite boundaries,
            completed by the leftover in the current commit region.
        
        Side effect:
        Update ``self.pairs`` with error strings ending
        at the future boundary of the commit region.
        """
        error_count, self.pairs = self._LOGICAL_COUNTER.count(self.pairs)
        return error_count
    
    def sim_cycles_given_weight(self, decoder, weight, n):
        raise NotImplementedError("Not implemented for streaming schemes.")


class Forward(_Streaming):
    """Forward decoding scheme. Also known as overlapping recovery method.
    
    Extends ``_Streaming``.
    """

    step_counts: list[int | tuple[int, int] | None]
    history: list[tuple[set[Edge], set[Edge], set[Node]]]
    """A list of tuples (error, leftover, artificial defects) for each cycle."""

    @staticmethod
    def __str__() -> str:
        return 'forward'

    @override
    def reset(self):
        super().reset()
        try: del self.history
        except: pass

    def _make_error(
            self,
            buffer_leftover: set[Edge],
            noise_level: float,
            exclude_future_boundary: bool = False,
    ):
        """Lower ``buffer_leftover`` by commit height
        and sample edges from freshly discovered region with probability ``noise_level``.
        
        
        :param buffer_leftover: the current error in the buffer region.
        :param noise_level: probability for an edge to bitflip.
        :param exclude_future_boundary: passed to ``self._CODE.make_error``.
        
        :return error: The set of bitflipped edges.
        """
        # lower `buffer_leftover` by commit height
        seen: set[Edge] = set()
        for e in buffer_leftover:
            seen.add(self._CODE.raise_edge(e, -self._COMMIT_HEIGHT))
        # populate freshly discovered region with new errors
        unseen = self._CODE.make_error(noise_level, exclude_future_boundary=exclude_future_boundary)
        return seen | unseen

    def _get_leftover(self, error: set[Edge], correction: set[Edge]):
        """Sequentially compose ``error`` and commit region of ``correction``.
        
        
        :param error: the set of bitflipped edges in the window.
        :param correction: the decoder output for the whole window.
        
        
        :return commit_leftover: The leftover, i.e. sequential composition of ``error`` and ``correction``, in the commit region.
        :return buffer_leftover: The part of ``error`` in the buffer region.
        """
        commit_leftover = error.intersection(self._COMMIT_EDGES) ^ correction.intersection(self._COMMIT_EDGES)
        buffer_leftover = error.intersection(self._BUFFER_EDGES)
        return commit_leftover, buffer_leftover

    @override
    def get_logical_error(self, commit_leftover: set[Edge]):
        """Count logical errors completed in current commit.
        
        Additional input over ``_Streaming.get_logical_error``:
        ``commit_leftover`` the leftover in the commit region.
        """
        for e in commit_leftover:
            self.pairs.load(e)
        return super().get_logical_error()

    def _get_syndrome(
            self,
            commit_leftover: set[Edge],
            error: set[Edge],
    ):
        """Get syndrome of ``error`` accounting for artificial defects due to previous commit."""
        artificial_defects = self._get_artificial_defects(commit_leftover)
        return self._CODE.get_syndrome(error) ^ artificial_defects

    def _get_artificial_defects(self, commit_leftover: set[Edge]):
        """Get artificial defects due to ``commit_leftover``.
        
        See Skoric et al. [arXiv:2209.08552v2, Section I B]
        for artificial defect definition.
        """
        commit_syndrome = self._CODE.get_syndrome(commit_leftover)
        return {self._CODE.raise_node(v, -self._COMMIT_HEIGHT) for v in commit_syndrome
                if v[self._CODE.TIME_AXIS] == self._COMMIT_HEIGHT}

    def _make_error_in_buffer_region(self, noise_level: float):
        """Sample edges from buffer region.
        
        
        :param noise_level: Characteristic probability if circuit-level noise;
            else, bitflip probability.
        
        
        :return error: The set of bitflipped edges in the buffer region.
            Each edge bitflips with probability defined by its multiplicity
            if circuit-level noise; else, probability ``noise_level``.
        """
        # TODO: test this method
        if self._COMMIT_HEIGHT >= self._BUFFER_HEIGHT:
            error = self._CODE.make_error(noise_level)
        else:
            rep_count = math.ceil(self._BUFFER_HEIGHT / self._COMMIT_HEIGHT)
            error: set[Edge] = set()
            for _ in itertools.repeat(None, rep_count):
                error = self._make_error(error, noise_level)
        return error & set(self._BUFFER_EDGES)
    
    @override
    def run(
            self,
            decoder: 'Decoder',
            noise_level: float,
            n: int,
            draw=False,
            log_history=False,
            **kwargs_for_draw_run,
    ):
        """Simulate ``n`` decoding cycles (in the steady state), for analysing accuracy and throughput.
        
        
        :param decoder: the decoder.
        :param noise_level: The noise level to simulate.
        :param n: is decoding cycle count in the steady state.
        
        :return failure_count: Failure count.
        :return slenderness: (total layer count / code distance).
        
        Side effect: Populate ``self.step_counts`` with ``n`` entries,
        each being the step count of one decoding cycle in the steady state.
        """
        log_history |= draw
        self.reset()
        if n < 1: raise ValueError("n must be positive integer.")
        m = 0
        commit_leftover: set[Edge] = set()
        buffer_leftover = self._make_error_in_buffer_region(noise_level)
        # `cleanse_count` additional decoding cycles ensures window is free of defects
        cleanse_count = self.WINDOW_HEIGHT // self._COMMIT_HEIGHT
        # If F is the last freshly discovered region of edges that can flip (that excludes a future boundary),
        # then `cleanse_count` is the number of decoding cycles needed for F to be fully in the commit region.
        # If F includes the future boundary, then need 1 more than this.
        if log_history:
            self.history = []
        for instantaneous_noise_level, time, exclude_future_boundary in itertools.chain(
            itertools.repeat((noise_level, True, False), n),
            itertools.repeat((noise_level, False, True), 1),
            itertools.repeat((0, False, False), cleanse_count),
        ):
            error = self._make_error(
                buffer_leftover,
                instantaneous_noise_level,
                exclude_future_boundary=exclude_future_boundary,
            )
            artificial_defects = self._get_artificial_defects(commit_leftover)
            syndrome = self._CODE.get_syndrome(error) ^ artificial_defects
            decoder.reset()
            step_count = decoder.decode(syndrome)
            if time:
                self.step_counts.append(step_count)
            commit_leftover, buffer_leftover = self._get_leftover(error, decoder.correction)
            if log_history:
                self.history.append((error, commit_leftover | buffer_leftover, artificial_defects))
            m += self.get_logical_error(commit_leftover)
        if draw:
            self._draw_run(**kwargs_for_draw_run)
        return m, (self._BUFFER_HEIGHT + (n+1)*self._COMMIT_HEIGHT) / self._CODE.D

    def _draw_run(
        self,
        fig_width: float | None = None,
        x_offset=constants.STREAM_X_OFFSET,
        subplot_hspace=-0.1,
    ):
        """Draw the history of ``self.run``."""
        import matplotlib.pyplot as plt
        column_count, row_count = len(self.history), 2
        if fig_width is None:
            fig_width = 1.5 if self._CODE.DIMENSION==2 else self._CODE.D*3/5
        plt.figure(figsize=(
            fig_width * column_count,
            fig_width * row_count * self.WINDOW_HEIGHT/self._CODE.D
        ))
        for k, (error, leftover, artificial_defects) in enumerate(self.history, start=1):
            for l, edges in enumerate((error, leftover)):
                plt.subplot(row_count, column_count, column_count*l + k)
                self._CODE.draw(
                    edges,
                    syndrome=self._CODE.get_syndrome(edges) ^ artificial_defects,
                    x_offset=x_offset,
                    with_labels=False,
                )
        plt.tight_layout()
        plt.subplots_adjust(hspace=subplot_hspace)


class Frugal(_Streaming):
    """Frugal decoding scheme.
    
    Extends ``_Streaming``.
    """

    @staticmethod
    def __str__() -> str:
        return 'frugal'

    def __init__( self, code, commit_height, buffer_height, commit_edges):
        super().__init__(code, commit_height, buffer_height, commit_edges)
        self.error: set[Edge] = set()
        """The set of flipped edges."""
        self._future_boundary_syndrome: set[Node] = set()
        """The set of defects in the future boundary of the viewing window."""
        self.step_counts: list[int]

    @override
    def reset(self):
        super().reset()
        self.error.clear()
        self._future_boundary_syndrome.clear()

    def advance(
            self,
            prob: float,
            decoder: 'Snowflake',
            exclude_future_boundary: bool = False,
            **kwargs,
    ) -> int:
        """Advance 1 decoding cycle.
        
        
        :param prob: instantaneous noise level.
        :param decoder: the frugal-compatible decoder to use.
            Currently, only Snowflake is supported.
        :param exclude_future_boundary: passed to ``self._CODE.make_error``.
        :param log_history: forwarded to ``decoder.decode``.
        :param time_only: forwarded to ``decoder.decode``.
        :param defects_possible: forwarded to ``decoder.decode``.
        
        :return timestep_count: number of decoder timesteps to complete decoding cycle.
        """
        self._raise_window()
        error = self._CODE.make_error(
            prob,
            exclude_future_boundary=exclude_future_boundary,
        )
        syndrome = self._load(error)
        return decoder.decode(syndrome, **kwargs)

    def _raise_window(self):
        """Raise window by ``self._COMMIT_HEIGHT`` layers."""
        # TODO: store ``lowness, next_edge`` as attributes of each edge.
        next_error: set[Edge] = set()
        for e in self.error:
            lowness = sum(v[self._CODE.TIME_AXIS] == 0 for v in e)
            if lowness == 0:
                next_edge = self._CODE.raise_edge(e, delta_t=-self._COMMIT_HEIGHT)
                next_error.add(next_edge)
            else:
                self.pairs.load(e)
        self.error = next_error

    def _load(self, error: set[Edge]):
        """Load incremental ``error``.
        
        :param error: The incremental error, which should never intersect ``self.error``.
        
        :return syndrome: The incremental syndrome due to ``error``.
        
        Side effect: Update ``self.error`` and ``self.future_boundary_syndrome``.
        """
        self.error |= error
        syndrome = {self._CODE.raise_node(v, delta_t=-self._COMMIT_HEIGHT)
            for v in self._future_boundary_syndrome}
        self._future_boundary_syndrome.clear()
        for e in error:
            for v in e:
                if not self.is_boundary(v):
                    syndrome.symmetric_difference_update({v})
                elif v[self._CODE.TIME_AXIS] == self.WINDOW_HEIGHT:  # in future boundary
                    self._future_boundary_syndrome.symmetric_difference_update({v})
        return syndrome

    @override
    def run(
            self,
            decoder: 'Snowflake',
            noise_level: float,
            n: int,
            draw: Literal[False, 'fine', 'coarse'] = False,
            log_history: Literal[False, 'fine', 'coarse'] = False,
            confidence_scores: Iterable[str] = (),
            time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
            **kwargs_for_draw_decode,
    ):
        """Simulate ``n*d`` decoding cycles (in the steady state).
        
        
        :param decoder: The decoder; currently, only Snowflake is supported.
        :param noise_level: The noise level to simulate.
        :param n: The number of decoding cycles in the steady state, divided by the code distance.
            For Snowflake, there is 1 decoding cycle per stabiliser measurement round.
        :param draw: Whether to draw.
        :param log_history: Whether to populate ``history`` attribute.
        :param confidence_scores: An iterable of DCSs to compute and record after each decoding cycle.
        :param time_only: Whether runtime includes a timestep
            for each drop, each grow, and each merging step ('all');
            each merging step only ('merging');
            or each unrooting step only ('unrooting').
            Note: if commit height is 1 and window height is d,
            then changing from 'merging' to 'all' simply increases each step count by 2d
            [for a total step count increase of 2d(n-1)]
            in the case of Snowflake with the 1:1 schedule.
            In the case of the 2:1 schedule, the increase of each step count is 3d.
            This can be done post-run via ``add_ignored_timesteps``.
        :param kwargs_for_draw_decode: passed to ``decoder.draw_decode``
            e.g. ``margins=(0.1, 0.1)``.
        
        :return failure_count: Failure count.
        :return slenderness: (total layer count) / (code distance).
        
        Side effect: Populate ``self.step_counts`` with ``n`` entries,
        each being the step count of ``d`` decoding cycles in the steady state.
        """
        d = self._CODE.D
        self.reset()
        if n < 1: raise ValueError("n must be positive integer.")
        decoder.reset()
        if draw:
            log_history = draw
        if log_history:
            decoder.init_history()
        m = 0
        transient_count = math.ceil(self.WINDOW_HEIGHT / self._COMMIT_HEIGHT)
        # require `transient_count` decoding cycles to reach steady state
        cleanse_count = 2 * self.WINDOW_HEIGHT
        # require `cleanse_count` decoding cycles after receiving last sheet of syndrome
        # to guarantee last defect is annihilated
        # TODO: tighten this bound
        # Note: cleanse_count = math.ceil(self.WINDOW_HEIGHT / self._COMMIT_HEIGHT) does NOT work
        # as the committed correction will be incomplete
        for instantaneous_noise_level, advance_count, time, exclude_future_boundary in itertools.chain(
            ((noise_level, transient_count, False, False),),
            itertools.repeat((noise_level, d, True, False), n),
            ((noise_level, d-1, False, False),),
            ((noise_level, 1, False, True),),
            ((0, cleanse_count, False, False),),
        ):
            step_count = 0
            for _ in itertools.repeat(None, advance_count):
                step_count += self.advance(
                    instantaneous_noise_level,
                    decoder,
                    exclude_future_boundary=exclude_future_boundary,
                    log_history=log_history,
                    confidence_scores=confidence_scores,
                    noise_level_for_priors=noise_level,
                    time_only=time_only,
                )
                m += self.get_logical_error()
            if time:
                self.step_counts.append(step_count)
        if draw:
            decoder.draw_decode(**kwargs_for_draw_decode)
        return m, transient_count / d + n + 1
    
    def sample_latency(
            self,
            decoder: 'Snowflake',
            noise_level: float,
            draw: Literal[False, 'fine', 'coarse'] = False,
            log_history: Literal[False, 'fine', 'coarse'] = False,
            time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
            **kwargs_for_draw_decode,
    ):
        """Sample the latency of the decoder in the frugal decoding scheme.
        
        This runs a memory experiment of ``self.WINDOW_HEIGHT+1`` measurement rounds
        i.e. ``self.WINDOW_HEIGHT+2`` sheets of syndrome data are produced.
        Assumes commit height is 1.
        
        Inputs same as in ``run``.
        
        
        :return latency: The number of timesteps
            from receiving the last measurement round to outputting the final correction.
        """
        self.reset()
        decoder.reset()
        if draw:
            log_history = draw
        if log_history:
            decoder.init_history()
        latency = 0
        defects_possible = True
        for instantaneous_noise_level, include_in_latency in itertools.chain(
            itertools.repeat((noise_level, False), self.WINDOW_HEIGHT+1),
            ((noise_level, True),),
            itertools.repeat((0, True), self.WINDOW_HEIGHT-1),
        ):
            if defects_possible and instantaneous_noise_level==0:
                defects_possible = bool(decoder.syndrome)
            step_count = self.advance(
                instantaneous_noise_level,
                decoder,
                exclude_future_boundary=include_in_latency,
                log_history=log_history,
                time_only=time_only,
                defects_possible=defects_possible,
            )
            if include_in_latency:
                latency += step_count
        if draw:
            decoder.draw_decode(**kwargs_for_draw_decode)
        return latency


from itertools import repeat
from unittest.mock import patch, call

from pandas import DataFrame, MultiIndex

from localuf.sim import runtime


def test_validate_only():
    stages = ('SV',)
    ds = [1, 2]
    noise_levels = [0, 1]
    n = 2
    call_count = len(ds) * len(noise_levels) * n
    with (
        patch('localuf.Surface.make_error', return_value=set()) as mock_me,
        patch('localuf.Surface.get_syndrome', return_value=set()) as mock_gs,
        patch('localuf.decoders.luf.main.LUF.validate', return_value=0) as mock_validate,
        patch('localuf.decoders.luf.main.LUF.reset') as mock_reset,
    ):
        data = runtime.batch(
            ds=ds,
            noise_levels=noise_levels,
            n=n,
            noise='code capacity',
        )
        assert mock_me.call_args_list == len(ds) * [
            call(noise_level) for noise_level in noise_levels for _ in repeat(None, times=n)
        ]
        assert mock_gs.call_args_list == call_count * [call(set())]
        assert mock_validate.call_args_list == call_count * [call(set())]
        assert mock_reset.call_args_list == call_count * [call()]
    assert type(data) is DataFrame
    assert data.shape == (n, len(ds) * len(noise_levels))
    assert data.columns.names == ['stage', 'd', 'p']
    assert (data.columns == MultiIndex.from_product([stages, ds, noise_levels])).all()


def test_full():
    stages = ('BP', 'SV')
    ds = [1, 2]
    noise_levels = [0, 1]
    n = 2
    call_count = len(ds) * len(noise_levels) * n
    with (
        patch('localuf.Surface.make_error', return_value=set()) as mock_me,
        patch('localuf.Surface.get_syndrome', return_value=set()) as mock_gs,
        patch('localuf.decoders.luf.main.LUF.decode', return_value=(0, 0)) as mock_decode,
        patch('localuf.decoders.luf.main.LUF.reset') as mock_reset,
    ):
        data = runtime.batch(
            ds=ds,
            noise_levels=noise_levels,
            n=n,
            noise='code capacity',
            validate_only=False,
        )
        assert mock_me.call_args_list == len(ds) * [
            call(noise_level) for noise_level in noise_levels for _ in repeat(None, times=n)
        ]
        assert mock_gs.call_args_list == call_count * [call(set())]
        assert mock_decode.call_args_list == call_count * [call(set())]
        assert mock_reset.call_args_list == call_count * [call()]
    assert type(data) is DataFrame
    assert data.shape == (n, len(stages) * len(ds) * len(noise_levels))
    assert data.columns.names == ['stage', 'd', 'p']
    assert (data.columns == MultiIndex.from_product([stages, ds, noise_levels])).all()


from localuf.decoders.luf import LUF, MacarNode, Nodes
from localuf.decoders.luf.constants import Stage


class AlternativeMacarNodes(Nodes):
    """Alternative to ``MacarNodes`` where ``valid``
    is an attribute rather than a property
    and is updated after every PRESYNCING stage.
    
    Extends ``Nodes``.
    Designed to emulate faster than ``MacarNodes`` but empirically not.
    Hence currently unused.
    """

    def __init__(self, luf: LUF) -> None:
        super().__init__(luf)
        self._dc = {v: MacarNode(nodes=self, index=v)
                    for v in luf.CODE.NODES}
        self.reset()

    @property
    def dc(self): return self._dc

    @property
    def busy(self):
        return any(node.busy for node in self.dc.values())

    def reset(self):
        self.valid = True
        return super().reset()

    def load(self, syndrome):
        super().load(syndrome)
        self.valid = False

    def update_valid(self):
        """Update ``valid`` attribute. Call after every PRESYNCING stage."""
        self.valid = not any(node.active for node in self.dc.values())

    def advance(self):
        super().advance()
        if self.LUF.CONTROLLER.stage is Stage.PRESYNCING:
            self.update_valid()


import pytest

from localuf.type_aliases import Node
from localuf.decoders import UF, BUF, NodeBUF
from localuf.decoders.uf import _Cluster
from localuf.constants import Growth
from localuf.type_aliases import Node

@pytest.fixture
def v00():
    return 0, 0

@pytest.fixture
def v000():
    return 0, 0, 0

@pytest.fixture
def syndrome7F() -> set[Node]:
    return {
        (0, 0),
        (0, 1),
        (2, 0),
        (0, 4),
        (1, 4),
        (3, 4),
        (4, 4),
        (4, 3),
    }

@pytest.fixture
def fixture_test_update_self_after_union():
    def f(uf_after_union: tuple[BUF | NodeBUF, _Cluster, _Cluster]):
        buf, larger, smaller = uf_after_union
        # check smaller cluster deleted
        assert smaller.root not in buf.clusters
        assert all(smaller not in bucket for bucket in buf.buckets.values())
        # check larger cluster in clusters but not in buckets
        assert larger.root in buf.clusters
        assert all(larger not in bucket for bucket in buf.buckets.values())
    return f

@pytest.fixture
def fixture_test_grow():
    def f(uf: UF):
        
        # setup
        sq = (2, 0)
        c = uf.clusters[sq]
        sq_incident_edges = {
            ((1, 0), sq),
            ((2, -1), sq),
            (sq, (2, 1)),
            (sq, (3, 0))
        }
        merge_ls, changed_edges = [], set()

        # test core functionality
        uf._grow(c, merge_ls, changed_edges)
        assert merge_ls == []
        assert changed_edges == sq_incident_edges
        assert all(uf.growth[e] is Growth.HALF for e in sq_incident_edges)
        return sq, c
    return f


@pytest.fixture
def uvw() -> tuple[Node, Node, Node]:
    u, v, w = (0, 0), (0, 1), (0, 2)
    return u, v, w


@pytest.fixture
def fixture_test_union():
    def f(uf_after_union: tuple[BUF | NodeBUF, _Cluster, _Cluster]):
        uf, larger, smaller = uf_after_union
        assert uf.parents[smaller.root] == larger.root
        assert larger.size == 2
        assert larger.odd
        assert larger.boundary == larger.root
    return f


import pytest

from localuf.decoders.snowflake.main import _Surface


def test_syndrome_vector_to_set():
    converter = _Surface(3, 2)
    assert converter._syndrome_vector_to_set('100100') == {(2, 0, 1), (1, 1, 1)}
    assert converter._syndrome_vector_to_set('010010') == {(2, 1, 1), (0, 0, 1)}
    assert converter._syndrome_vector_to_set('001001') == {(1, 0, 1), (0, 1, 1)}
    assert converter._syndrome_vector_to_set('000000') == set()
    with pytest.raises(ValueError):
        converter._syndrome_vector_to_set('10010')


@pytest.mark.parametrize("window_height", [1, 2, 3])
def test_bit_position_to_defect(window_height: int):
    converter = _Surface(3, window_height)
    assert converter._bit_position_to_defect(0) == (2, 0, window_height-1)
    assert converter._bit_position_to_defect(1) == (2, 1, window_height-1)
    assert converter._bit_position_to_defect(2) == (1, 0, window_height-1)
    assert converter._bit_position_to_defect(3) == (1, 1, window_height-1)
    assert converter._bit_position_to_defect(4) == (0, 0, window_height-1)
    assert converter._bit_position_to_defect(5) == (0, 1, window_height-1)


@pytest.mark.parametrize("window_height", [1, 2, 3])
def test_defect_to_bit_position(window_height: int):
    converter = _Surface(3, window_height)
    assert converter._defect_to_bit_position((2, 0, window_height-1)) == 0
    assert converter._defect_to_bit_position((2, 1, window_height-1)) == 1
    assert converter._defect_to_bit_position((1, 0, window_height-1)) == 2
    assert converter._defect_to_bit_position((1, 1, window_height-1)) == 3
    assert converter._defect_to_bit_position((0, 0, window_height-1)) == 4
    assert converter._defect_to_bit_position((0, 1, window_height-1)) == 5
    with pytest.raises(ValueError):
        converter._defect_to_bit_position((3, 0, window_height))


@pytest.mark.parametrize("window_height", [2, 3])
def test_set_to_syndrome_vector(window_height: int):
    converter = _Surface(3, window_height)
    assert converter._set_to_syndrome_vector({(2, 0, window_height-1), (1, 1, window_height-1)}) == '100100'
    assert converter._set_to_syndrome_vector({(2, 1, window_height-1), (0, 0, window_height-1)}) == '010010'
    assert converter._set_to_syndrome_vector({(1, 0, window_height-1), (0, 1, window_height-1)}) == '001001'
    assert converter._set_to_syndrome_vector(set()) == '000000'


from unittest import mock

from localuf.decoders.snowflake import _Node


def test_find_broken_pointers(sfn3: _Node):
    for pointer in ('D', 'SD', 'WD', 'NWD'):
        with mock.patch("localuf.decoders.snowflake.main._FullUnrooter.start") as mock_start:
            sfn3.pointer = pointer
            sfn3.FRIENDSHIP.find_broken_pointers()
            mock_start.assert_called_once()


"""Unit test the `Controller` class.

These tests use a 2D distance-3 controller fixture.
Need not repeat on a 3D fixture.
"""
# from unittest import mock

# import pytest

# from localuf.decoders.luf import LUF, Controller


import numpy as np
import pytest

from localuf import Surface
from localuf.noise import CircuitLevel
from localuf.type_aliases import Edge, FourInts


@pytest.mark.parametrize('d', range(3, 7, 2), ids=lambda d: f'd{d}')
def test_init(d: int):

    sf1 = Surface(d, 'circuit-level', _merge_redundant_edges=False)
    sf2 = Surface(d, 'circuit-level')
    sf2_edges = set(sf2.EDGES)

    # manually merge edges of sf1
    dc: dict[Edge, list[FourInts]] = {}
    model_1: CircuitLevel = sf1.NOISE  # type: ignore
    for m, es in model_1._FRESH_EDGES.items():
        for e in es:
            if e not in sf2_edges:
                sub = sf1._substitute(e)
                if sub in dc:
                    dc[sub].append(m)
                else:
                    dc[sub] = [m]
            else:
                if e in dc:
                    dc[e].append(m)
                else:
                    dc[e] = [m]
    edges = {e: sum(np.array(m) for m in ls) for e, ls in dc.items()}

    model_2: CircuitLevel = sf2.NOISE  # type: ignore
    for e, m in edges.items():
        key: FourInts = tuple(m) # type: ignore
        assert e in model_2._FRESH_EDGES[key]


def test_EDGES(toy_cl: CircuitLevel, e_westmost: tuple[Edge, Edge]):
    e0, e1 = e_westmost
    assert len(toy_cl._FRESH_EDGES) == 2
    assert toy_cl._FRESH_EDGES[1, 0, 2, 0] == [e0]
    assert toy_cl._FRESH_EDGES[5, 0, 3, 0] == [e1]


import pytest

from localuf.codes import Surface
from localuf.decoders.uf import UF
from localuf.decoders.node_uf import NodeUF


@pytest.mark.parametrize('noise', ('phenomenological', 'circuit-level'))
@pytest.mark.parametrize('decoder_class', (UF, NodeUF))
@pytest.mark.parametrize('row', (0, 1))
class TestUpdateBoundary:

    def test_d2(self, noise, decoder_class, row):
        code = Surface(2, noise)
        decoder = decoder_class(code, inclination='west')
        syndrome = {(row, 0, 0)}
        decoder.decode(syndrome)
        assert decoder.CODE.get_logical_error(decoder.correction)
        assert decoder.correction == {((row, -1, 0), (row, 0, 0))}

    def test_d4(self, noise, decoder_class, row):
        code = Surface(4, noise)
        decoder = decoder_class(code, inclination='west')
        syndrome = {(row, 1, 0)}
        decoder.decode(syndrome)
        assert decoder.CODE.get_logical_error(decoder.correction)
        assert decoder.correction == {
            ((row, -1, 0), (row, 0, 0)),
            ((row, 0, 0), (row, 1, 0)),
        }


"""Constants for Snowflake decoder."""

from enum import IntEnum

RESET = -1
"""`cid` value for nodes to be unrooted."""

class Stage(IntEnum):
    """Constants for stage values."""

    INCREMENT = 1
    STAGE_COUNT = 5
    """Stage count."""
    DROP, GROW_WHOLE, MERGING_WHOLE, GROW_HALF, MERGING_HALF = range(STAGE_COUNT)

    def __iadd__(self, other: 'Stage'):
        return Stage(self.value + other.value)

    def __imod__(self, other: 'Stage'):
        return Stage(self.value % other.value)

    def __str__(self):
        match self:
            case Stage.DROP:          return 'D'
            case Stage.GROW_WHOLE:    return 'G'
            case Stage.MERGING_WHOLE: return 'M'
            case Stage.GROW_HALF:     return 'g'
            case Stage.MERGING_HALF:  return 'm'


from localuf.decoders.luf import ActisNodes

def test_busy_signal_causality(actis5Fu_nodes: ActisNodes):
    """Ensure `busy_signal` of node whose ID is 0 resets not countdown of
    `ActisNodes` in same timestep when `ActisNodes.advance()` called.
    """
    rep = actis5Fu_nodes.dc[0, -1]
    rep.busy_signal = True
    rep.FRIENDSHIP.relay_signals()

    assert not actis5Fu_nodes.busy_signal
    assert actis5Fu_nodes.next_busy_signal

    actis5Fu_nodes._update_unphysicals_for_actis()

    assert actis5Fu_nodes.busy_signal
    assert not actis5Fu_nodes.next_busy_signal


from typing import Iterable

import pytest

from localuf.noise import CircuitLevel
from localuf.type_aliases import EdgeType, MultiplicityVector

@pytest.fixture
def diagonals():
    return (
        'SD',
        'EU west corners',
        'EU east corners',
        'EU edge',
        'EU centre',
        'SEU',
    )


@pytest.fixture
def assert_these_multiplicities_unchanged():
    def f(
            m: dict[EdgeType, MultiplicityVector],
            edge_types: Iterable[EdgeType],
    ):
        for et in edge_types:
            assert tuple(m[et]) \
                == CircuitLevel._DEFAULT_MULTIPLICITIES[et]
    return f


@pytest.fixture
def ftto():
    return (4, 3, 2, 1)


from localuf.decoders.node_uf import _NodeCluster, NodeUF
from localuf.type_aliases import Node


class NodeBUF(NodeUF):
    """The graph used by Bucket UF decoder in Huang2020.
    
    Extends ``decoders.NodeUF``.
    
    Overriden attributes:
    
    * ``buckets`` replaces ``active_clusters``
        and is a dictionary where each key a cluster size;
        value, a set of all active clusters of that size.
    
    Overriden methods:
    
    * ``reset``.
    * ``load``.
    * ``validate_syndrome``.
    * ``_update_self_after_union``.
    """

    def reset(self):
        super().reset()
        d = self.CODE.D
        n_nodes = (d+1) * d
        self.buckets: dict[int, set[_NodeCluster]] = {size: set() for size in range(1, n_nodes+1)}

    def load(self, syndrome: set[Node]):
        super().load(syndrome)
        for cluster in self.active_clusters:
            self.buckets[cluster.size].add(cluster)
        del self.active_clusters

    # SYNDROME VALIDATION

    def validate(
        self,
        syndrome,
        log_history=False,
    ):
        """Grow clusters (always smallest first) until they are all valid."""
        self.load(syndrome)
        if log_history: self.init_history()
        for bucket in self.buckets.values():
            while bucket:
                self._growth_round(log_history, clusters_to_grow=bucket)

    def _update_self_after_union(self, larger: _NodeCluster, smaller: _NodeCluster, old_larger_size: int):
        """Update attributes clusters and buckets after union of larger w/ smaller."""

        # DELETE SMALLER CLUSTER
        del self.clusters[smaller.root]
        # discard instead of remove as smaller may be a node not in syndrome
        self.buckets[smaller.size].discard(smaller)

        # UPDATE buckets DICTIONARY
        # discard instead of remove as larger could have had boundary before union, hence was not active
        self.buckets[old_larger_size].discard(larger)
        if larger.odd and not larger.boundary:
            self.buckets[larger.size].add(larger)


"""Module for decoders.

Available decoders:

* Actis
* BUF
* Macar
* MWPM
* NodeBUF
* NodeUF
* Snowflake
* UF

In drawings of UF-type decoders:

* ungrown edges are invisible
* half-grown edges are dotted
* fully grown edges are solid
"""

from localuf.decoders.uf import UF
from localuf.decoders.buf import BUF
from localuf.decoders.node_uf import NodeUF
from localuf.decoders.node_buf import NodeBUF
from localuf.decoders.luf import Macar, Actis
from localuf.decoders.mwpm import MWPM
from localuf.decoders.snowflake import Snowflake


from typing import Literal, TYPE_CHECKING

import networkx as nx

from localuf.constants import Growth
from localuf.type_aliases import Node, Edge

if TYPE_CHECKING:
    from localuf.decoders._base_uf import BaseUF
    from localuf.decoders.luf.main import MacarNode, ActisNode, _Node

class _PolicyMixin:
    """Provides ``_NODES`` and ``_growth`` attributes."""

    def __init__(
            self,
            nodes: 'dict[Node, MacarNode] | dict[Node, ActisNode]',
            growth: dict[Edge, Growth],
    ):
        self._NODES = nodes
        self._growth = growth


class DigraphMaker(_PolicyMixin):
    """Maker of NetworkX digraph.
    
    Instance attributes:
    
    * ``_NODES`` dictionary of nodes.
    * ``_growth`` dictionary of growths.
    """

    @property
    def pointer_digraph(self):
        """Return a NetworkX digraph representing the fully grown edges used by pointers,
        the set of its edges as directed edges,
        the set of its edges as undirected edges.
        """
        dig = nx.DiGraph()
        dig.add_nodes_from(self._NODES.keys())
        dig_diedges: list[Edge] = []
        dig_edges: list[Edge] = []
        for u, node in self._NODES.items():
            if node.pointer != 'C':
                e, index = node.NEIGHBORS[node.pointer]
                if self._growth[e] is Growth.FULL:
                    v = e[index]
                    dig.add_edge(u, v)
                    dig_diedges.append((u, v))
                    dig_edges.append(e)
        return dig, dig_diedges, dig_edges
    

class DecodeDrawer:
    """Provides ``draw_decode``."""

    def __init__(
            self,
            fig_width: float,
            fig_height: None | float = None,
    ):
        self._FIG_WIDTH = fig_width
        self._FIG_HEIGHT = fig_width if fig_height is None else fig_height

    def draw(
            self,
            history: list['BaseUF'],
            style: Literal['interactive', 'horizontal', 'vertical'] = 'interactive',
            fig_width: None | float = None,
            fig_height: None | float = None,
            **kwargs_for_networkx_draw,
    ):
        """Draw the decoder's history."""
        from matplotlib import pyplot as plt
        if fig_width is None: fig_width = self._FIG_WIDTH
        if fig_height is None: fig_height = self._FIG_HEIGHT
        n_plots = len(history)
        if style == 'interactive':
            from ipywidgets import interact, BoundedIntText
            @interact(timestep=BoundedIntText(
                value=1,
                min=1,
                max=n_plots,
                step=1,
                description="timestep:",
                disabled=False,
            ))
            def f(timestep: int):
                plt.figure(figsize=(fig_width, fig_height))
                history[timestep-1].draw_growth(**kwargs_for_networkx_draw)
        elif style == 'horizontal':
            plt.figure(figsize=(fig_width*n_plots, fig_height))
            for k, older_self in enumerate(history, start=1):
                plt.subplot(1, n_plots, k)
                older_self.draw_growth(**kwargs_for_networkx_draw)
            plt.tight_layout()
        else:
            plt.figure(figsize=(fig_width, fig_height*n_plots))
            for k, older_self in enumerate(history, start=1):
                plt.subplot(n_plots, 1, k)
                older_self.draw_growth(**kwargs_for_networkx_draw)
            plt.tight_layout()


class AccessUpdater(_PolicyMixin):
    """Updater of access.
    
    Currently unused!
    """

    def update(self, node: '_Node'):
        node.access = {
            pointer: self._NODES[e[index]]
            for pointer, (e, index) in node.NEIGHBORS.items()
            if self._growth[e] is Growth.FULL
        }


from localuf import Repetition


def test_is_boundary(rp_forward: Repetition):
    n_boundaries = sum(rp_forward.is_boundary(v) for v in rp_forward.NODES)
    assert n_boundaries == 2 * rp_forward.SCHEME.WINDOW_HEIGHT + rp_forward.D - 1


from localuf.decoders import Snowflake
from tests.decoders.snowflake.Snowflake.test_Snowflake_unit import tds


from unittest import mock


def test_grow(snowflake_one_one: Snowflake):
    node_calls = [mock.call()] * len(snowflake_one_one.NODES)
    with (
        mock.patch(f"{tds}._Node.grow") as mock_growing,
        mock.patch(f"{tds}._Node.update_access") as mock_ua,
    ):
        snowflake_one_one._SCHEDULE.grow()
        assert mock_growing.call_args_list == node_calls
        assert mock_ua.call_args_list == node_calls


import pytest

from localuf.codes import Surface
from localuf.decoders.uf import _DefaultInclination


@pytest.fixture
def default_inclination(sf: Surface):
    return _DefaultInclination(sf.LONG_AXIS)


def test_update_boundary(two_clusters, default_inclination: _DefaultInclination):
    larger, smaller = two_clusters
    east, west = larger.root, smaller.root
    
    larger.boundary = None
    smaller.boundary = None
    default_inclination.update_boundary(larger, smaller)
    assert larger.boundary is None
    assert smaller.boundary is None
    
    larger.boundary = east
    default_inclination.update_boundary(larger, smaller)
    assert larger.boundary == east
    assert smaller.boundary is None
    
    larger.boundary = None
    smaller.boundary = west
    default_inclination.update_boundary(larger, smaller)
    assert larger.boundary == west
    assert smaller.boundary == west

    larger.boundary = east
    default_inclination.update_boundary(larger, smaller)
    assert larger.boundary == east
    assert smaller.boundary == west


import itertools

import pytest

from localuf import Surface
from localuf.noise import CodeCapacity
from localuf._schemes import Forward

def test_N_EDGES_attribute(sf: Surface):
    assert type(sf.N_EDGES) is int
    assert sf.N_EDGES == len(sf.EDGES)

def test_EDGES_attribute(sf: Surface):
    assert type(sf.EDGES) is tuple


class TestNODES:


    @pytest.fixture(
            name="sf_forward",
            params=range(3, 11, 2),
            ids=lambda x: f"d{x}",
    )
    def _sf_forward(self, request):
        return Surface(
            d=request.param,
            noise='phenomenological',
            scheme='forward',
    )

    def test_batch(self, sf: Surface):
        assert type(sf.NODES) is tuple
        if isinstance(sf.NOISE, CodeCapacity):
            assert len(sf.NODES) == sf.D * (sf.D+1)
        else:  # Phenomenological | CircuitLevel
            assert len(sf.NODES) == sf.D**2 * (sf.D+1)


    def test_forward(self, sf_forward: Surface):
        assert type(sf_forward.NODES) is tuple
        assert len(sf_forward.NODES) == sf_forward.D * (
            (sf_forward.D+1) * sf_forward.SCHEME.WINDOW_HEIGHT
            + sf_forward.D-1
        )

def test_LONG_AXIS_attribute(sf: Surface):
    assert sf.LONG_AXIS == 1

def test_DIMENSION_attribute(sf: Surface):
    assert sf.DIMENSION == len(sf.NODES[0])

def test_index_to_label(sf5F: Surface):
    assert sf5F.index_to_label((1, 1)) == 8
    # assert sf3.index_to_label((1, 0, 1)) == 37

def test_label_to_index(sf5F: Surface):
    assert sf5F.label_to_index(8) == (1, 1)

@pytest.mark.parametrize("sf", [
    "sf3F",
    "sf5F",
    "sf7F",
    "sf9F",
])
def test_index_label_consistency(sf: Surface, request):
    sf = request.getfixturevalue(sf)
    d = sf.D
    for a in range(d*(d+1)):
        assert sf.index_to_label(sf.label_to_index(a)) == a

def test_get_pos(sf3F: Surface):
    pos = sf3F.get_pos()
    assert type(pos) is dict


class TestIndexToId:


    @pytest.fixture
    def helper(self):
        def f(sf: Surface):
            d = sf.D
            boundary_IDs: set[int] = set()
            detector_IDs: set[int] = set()
            for v in sf.NODES:
                id_ = sf.index_to_id(v)
                if sf.is_boundary(v):
                    boundary_IDs.add(id_)
                else:
                    detector_IDs.add(id_)
            assert max(boundary_IDs) < min(detector_IDs)
            return d, boundary_IDs, detector_IDs
        return f


    def test_2D(self, sf5F: Surface, helper):
        d, boundary_IDs, detector_IDs = helper(sf5F)
        assert len(boundary_IDs) + len(detector_IDs) == d * (d+1)
        # highest ID boundary node
        assert sf5F.index_to_id((d-1, d-1)) == 2*d - 1
        # highest ID detector
        assert sf5F.index_to_id((d-1, d-2)) == d * (d+1) - 1


    def test_3D(self, sf5T: Surface, helper):
        d, boundary_IDs, detector_IDs = helper(sf5T)
        assert len(boundary_IDs) + len(detector_IDs) == d**2 * (d+1)
        assert sf5T.index_to_id((d-1, d-1, d-1)) == 2 * d**2 - 1
        assert sf5T.index_to_id((d-1, d-2, d-1)) == d**2 * (d+1) - 1


class TestCodeCapacityEdges:
    
    def test_correct_count(self, sf3F: Surface):
        count, edges = sf3F._code_capacity_edges(False)
        assert count == len(edges)
        nodes = set().union(*edges)
        assert nodes == set(itertools.product(range(3), range(-1, 3)))

    def test_merge_equivalent_boundary_nodes(self, sf3F: Surface):
        count, edges = sf3F._code_capacity_edges(True)
        assert count == len(edges)
        nodes = set().union(*edges)
        assert nodes == set(itertools.product(range(3), range(2))) | {
            (0, -1), (0, 2),
        }


class TestPhenomenologicalEdges:
    
    def test_correct_count(self, sf3T: Surface):
        count, edges = sf3T._phenomenological_edges(
            3,
            False,
            merge_equivalent_boundary_nodes=False,
        )
        assert count == len(edges)
        nodes = set().union(*edges)
        assert nodes == set(itertools.product(range(3), range(-1, 3), range(3)))

    def test_merge_equivalent_boundary_nodes(self, sf3T: Surface):
        count, edges = sf3T._phenomenological_edges(
            3,
            False,
            merge_equivalent_boundary_nodes=True,
        )
        assert count == len(edges)
        nodes = set().union(*edges)
        assert nodes == set(itertools.product(range(3), range(2), range(3))) | {
            (0, -1, 0), (0, 2, 0),
        }


class TestCircuitLevelEdges:


    def helper(
            self,
            surface: Surface,
            future_boundary: bool,
            merge_equivalent_boundary_nodes=False,
    ):
        """Output:
        * `d` code distance
        * `layer_count` number of layers in decoding graph
        * `edge_dict` maps from edge type
        (i.e. orientation and location)
        to tuple of all edges of that type.
        Always includes redundant edges.
        Inter-key order matters as used by `noise.forcers.ForceByEdge.force_error`.
        """
        d, h = surface.D, surface.SCHEME.WINDOW_HEIGHT
        layer_count = h if future_boundary else h-1
        len_merges = 2 * layer_count * (2*d-1)
        horizontal_edge_count = h * surface.DATA_QUBIT_COUNT
        u_per_layer = d*(d-1)
        sd_per_layer = (d-1)**2

        n_edges, edges, edge_dict, merges = surface._circuit_level_edges(
            h=h,
            future_boundary=future_boundary,
            _merge_redundant_edges=True,
            merge_equivalent_boundary_nodes=merge_equivalent_boundary_nodes,
        )

        eu_per_layer = d*(d-2)
        seu_per_layer = (d-1)*(d-2)
        timelike_edge_count = layer_count * (u_per_layer + sd_per_layer + eu_per_layer + seu_per_layer)
        assert n_edges == horizontal_edge_count + timelike_edge_count
        assert surface.N_EDGES == n_edges
        assert type(edges) is tuple
        assert len(edges) == n_edges
        assert len(set(edges)) == n_edges  # test uniqueness

        assert type(edge_dict) is dict
        assert len(edge_dict) == 12
        assert sum(len(es) for es in edge_dict.values()) == surface.N_EDGES + len_merges

        assert type(merges) is dict
        assert len(merges) == len_merges

        if not merge_equivalent_boundary_nodes:
            assert set(edges) & set(merges.keys()) == set()

            n_edges, edges, edge_dict_False, merges = surface._circuit_level_edges(
                h=h,
                future_boundary=future_boundary,
                _merge_redundant_edges=False,
            )

            eu_per_layer = d**2
            seu_per_layer = d*(d-1)
            timelike_edge_count = layer_count * (u_per_layer + sd_per_layer + eu_per_layer + seu_per_layer)
            assert n_edges == horizontal_edge_count + timelike_edge_count
            assert n_edges == surface.N_EDGES + len_merges
            assert len(edges) == n_edges
            assert len(set(edges)) == n_edges
            assert merges is None
            assert edge_dict_False == edge_dict

        return d, layer_count, edge_dict


    def test_future_boundary_False(self, sfCL: Surface):
        d, layer_count, edge_dict = self.helper(sfCL, False)

        eu_edge_NS = tuple(((i, j, t), (i, j+1, t+1)) for i in ( 0, d-1) for j in range(   d-2) for t in range(layer_count))
        eu_edge_EW = tuple(((i, j, t), (i, j+1, t+1)) for j in (-1, d-2) for i in range(1, d-1) for t in range(layer_count))
        assert edge_dict['EU edge'] == (*eu_edge_NS, *eu_edge_EW)

        seu_bulk     = tuple(((i, j, t), (i+1, j+1, t+1)) for i in range(d-1) for j in range(d-2) for t in range(layer_count))
        seu_boundary = tuple(((i, j, t), (i+1, j+1, t+1)) for j in  (-1, d-2) for i in range(d-1) for t in range(layer_count))
        assert edge_dict['SEU'] == (*seu_boundary, *seu_bulk)
        assert set(edge_dict['SEU']) == {
            ((i, j, t), (i+1, j+1, t+1))
            for i in range(d-1)
            for j in range(-1, d-1)
            for t in range(layer_count)
        }
        assert len(edge_dict['SEU']) == (d-1) * d * layer_count


    def test_future_boundary_True(self, sfCL_OL: Surface):
        d, layer_count, edge_dict = self.helper(sfCL_OL, True)

        eu_edge_NS = tuple(((i, j, t), (i, j+1, t+1)) for i in ( 0, d-1) for j in range(   d-2) for t in range(layer_count))
        eu_edge_W = tuple(((i,  -1, t), (i,   0, t+1)) for i in range(1, d-1) for t in range(-1, layer_count-1))
        eu_edge_E = tuple(((i, d-2, t), (i, d-1, t+1)) for i in range(1, d-1) for t in range(    layer_count  ))
        assert edge_dict['EU edge'] == (*eu_edge_NS, *eu_edge_W, *eu_edge_E)

        seu_bulk = tuple(((i, j, t), (i+1, j+1, t+1)) for i in range(d-1) for j in range(d-2) for t in range(layer_count))
        seu_W = tuple(((i, -1, t), (i+1, 0, t+1)) for i in range(d-1) for t in range(-1, layer_count-1))
        seu_E = tuple(((i, d-2, t), (i+1, d-1, t+1)) for i in range(d-1) for t in range(layer_count))
        assert edge_dict['SEU'] == (*seu_W, *seu_E, *seu_bulk)
        assert len(edge_dict['SEU']) == (d-1) * d * layer_count


    @pytest.mark.parametrize("buffer_height", range(1, 4), ids=lambda x: f"buffer_height {x}")
    @pytest.mark.parametrize("_merge_redundant_edges", [False, True], ids=lambda x: f"_merge_redundant_edges {x}")
    def test_t_start(self, sfCL_OL_scheme: Forward, buffer_height, _merge_redundant_edges):
        code = sfCL_OL_scheme._CODE
        commit_height = sfCL_OL_scheme._COMMIT_HEIGHT
        n_commit_edges, commit_edges, commit_edge_dict, commit_merges = code._circuit_level_edges(
            h=commit_height,
            future_boundary=True,
            _merge_redundant_edges=_merge_redundant_edges,
        )
        n_fresh_edges, fresh_edges, fresh_edge_dict, fresh_merges = code._circuit_level_edges(
            h=commit_height,
            future_boundary=True,
            _merge_redundant_edges=_merge_redundant_edges,
            t_start=buffer_height,
        )
        assert n_fresh_edges == n_commit_edges
        assert fresh_edges == tuple(code.raise_edge(e, buffer_height) for e in commit_edges)
        assert fresh_edge_dict == {edge_type: tuple(
            code.raise_edge(e, buffer_height) for e in edges
        ) for edge_type, edges in commit_edge_dict.items()}
        if _merge_redundant_edges:
            assert fresh_merges == {
                code.raise_edge(e, buffer_height):
                code.raise_edge(substitute, buffer_height)
                for e, substitute in commit_merges.items() # type: ignore
            }
        else:
            assert fresh_merges is None
            assert commit_merges is None

    
    def test_merge_equivalent_boundary_nodes(self, sfCL: Surface):
        self.helper(sfCL, False, True)


def test_substitute(sf3F: Surface, sf3T: Surface):
    assert sf3F._substitute(((0, -1), (1, 0))) == ((1, -1), (1, 0))
    assert sf3F._substitute(((2, 1), (1, 2))) == ((2, 1), ((2, 2)))
    assert sf3T._substitute(((0, -1, 0), (1, 0, 1))) == ((1, -1, 1), (1, 0, 1))
    assert sf3T._substitute(((2, 1, 1), (2, 2, 2))) == ((2, 1, 1), (2, 2, 1))


from unittest import mock

import pytest

from localuf import Surface
from localuf.decoders import UF
from localuf._determinants import SpaceDeterminant
from localuf._schemes import Batch

def test_DETERMINANT_attribute(batch3F: Batch):
    assert type(batch3F._DETERMINANT) is SpaceDeterminant

class TestRun:

    @pytest.mark.parametrize("n", range(1, 4))
    def test_code_capacity(self, sf3F: Surface, n: int):
        decoder = UF(sf3F)
        with mock.patch("localuf._schemes.Batch._sim_cycle_given_p", return_value=1) as mock_scgp:
            m, slenderness = sf3F.SCHEME.run(decoder, 1, n)
            assert mock_scgp.call_args_list == [mock.call(decoder, 1)] * n
            assert m == n
            assert slenderness == n

    @pytest.mark.parametrize("n", range(1, 4))
    def test_phenomenological(self, sf3T: Surface, n: int):
        decoder = UF(sf3T)
        fake_height = 2
        sf3T.SCHEME._WINDOW_HEIGHT = fake_height # type: ignore
        with mock.patch("localuf._schemes.Batch._sim_cycle_given_p", return_value=1) as mock_scgp:
            m, slenderness = sf3T.SCHEME.run(decoder, 1, n)
            assert mock_scgp.call_args_list == [mock.call(decoder, 1)] * n
            assert m == n
            assert slenderness == fake_height * n / 3


"""Classes for ``noise.CircuitLevel.force_error``.

This method samples an error from the set of all errors of a given weight
where weight can be an integer or a vector of integers
depending on the noise model.
"""


import abc
import itertools
import random
from functools import cached_property
from typing import Iterable

import numpy as np
from scipy.stats import binom

from localuf.type_aliases import Edge, FourInts, FourFloats


class _BaseForcer(abc.ABC):
    """Abstract base class for all forcers.
    
    Instance attributes:
    
    * ``_EDGES`` maps each multiplicity 4-vector
        to the corresponding subset as a list of edges.
    """

    def __init__(self, edges: dict[FourInts, list[Edge]]) -> None:
        self._EDGES = edges

    @abc.abstractmethod
    def force_error(self, weight: tuple[int, ...]) -> set[Edge]:
        """Make error whose weight in pair/edge subset ``k`` is ``weight[k]``."""

    @property
    @abc.abstractmethod
    def ALL_WEIGHTS(self) -> tuple[tuple[int, ...], ...]:
        """All possible ``force_error`` inputs."""
    
    @abc.abstractmethod
    def subset_probability(
        self,
        weights: tuple[tuple[int, ...], ...],
        pi: FourFloats,
    ) -> Iterable[float]:
        """Compute probability of each subset characterized by ``weight`` in ``weights``.
        
        See ``Noise.subset_probability``.
        
        
        :param weights: a tuple of inputs to ``force_error``.
        :param pi: a 4-tuple of probabilities.
        
        
        :returns: ``probs`` an iterable of probabilities where each corresponds to a ``weight`` in ``weights``.
        
        Example for ``ForceByPairBalanced``:
        
        * ``weights = ((0, 0), (0, 1), (1, 0), (1, 1))``
        * ``B_k = B(n=len(self.PAIR_POPULATIONS[k]), p=pi[k])``
        * ``probs`` = (
        
        ``B_0(0) * B_1(0)``,
        
        ``B_0(0) * B_1(1)``,
        
        ``B_0(1) * B_1(0)``,
        
        ``B_0(1) * B_1(1)``,
        )
        """


class _BaseForceByPair(_BaseForcer):
    """Base class for forcers whose error subsets are distinguished by the weight of each pair type.
    
    Extends ``_BaseForcer``.
    """

    PAIR_POPULATIONS: tuple[list[Edge], ...]
    """Population for each pair type."""

    def force_error(self, weight: tuple[int, ...]):
        """Make error whose weight in pair subset ``k`` is ``weight[k]``."""
        error: set[Edge] = set()
        for pairs, w in zip(self.PAIR_POPULATIONS, weight, strict=True):
            sample = random.sample(pairs, w)
            for e in sample:
                if e in error:
                    error.remove(e)
                else:
                    error.add(e)
        return error
    
    @property
    def ALL_WEIGHTS(self) -> tuple[tuple[int, ...], ...]:
        return tuple(itertools.product(*(
            range(len(pairs)+1) for pairs in self.PAIR_POPULATIONS
        )))
    
    def subset_probability(self, weights, pi) -> Iterable[float]:
        probs = np.ones(len(weights))
        weights = np.array(weights).T
        # truncate pi to same length as PAIR_POPULATIONS
        pairs_pi = zip(self.PAIR_POPULATIONS, pi)
        for row, (pairs, p) in zip(weights, pairs_pi, strict=True):
            probs *= binom.pmf(
                k=row,
                n=len(pairs),
                p=p,
            )
        return probs

    @property
    def _PAIR_POPULATIONS(self):
        """Helper property for ``PAIR_POPULATIONS``."""
        populations: list[list[Edge]] = []
        for k in range(4):
            pairs = []
            for m, edges in self._EDGES.items():
                pairs += m[k] * edges
            populations.append(pairs)
        return tuple(populations)


class ForceByPair(_BaseForceByPair):
    """Forcer whose error subsets are distinguished by the weight of each of the 4 pair types.
    
    Extends ``_BaseForceByPair``.
    Use when ``parametrization != 'balanced'``.
    In this class, each weight is a 4-tuple of integers.
    """
    
    @cached_property
    def PAIR_POPULATIONS(self):
        return self._PAIR_POPULATIONS


class ForceByPairBalanced(_BaseForceByPair):
    """Same as ``ForceByPair`` but aggregating the last 3 pair subsets.
    
    Extends ``_BaseForceByPair``.
    Use when ``parametrization == 'balanced'``.
    In this class, each weight is a 2-tuple of integers.
    """
    
    @cached_property
    def PAIR_POPULATIONS(self):
        p0, *p_rest = self._PAIR_POPULATIONS
        return (p0, sum(p_rest, start=[]))


class ForceByEdge(_BaseForcer):
    """Forcer whose error subsets are distinguished by the weight of each of the 13 edge types.
    
    Slower than ``ForceByPair`` as there are 13 edge types but only 4 pair types.
    Hence, use is discouraged.
    """

    def force_error(self, weight: tuple[int, ...]):
        """Make error whose weight in edge subset ``k`` is ``weight[k]``."""
        error: set[Edge] = set()
        try:
            for subset, w in zip(self._EDGES.values(), weight, strict=True):
                error.update(random.sample(subset, w))
        except ValueError as e:
            if str(e) == 'Sample larger than population or is negative':
                raise e
            else:
                raise ValueError(
                    f'len(weight)={len(weight)} != {len(self._EDGES)}=number of edge subsets'
                ) from e
        return error
    
    @property
    def ALL_WEIGHTS(self) -> tuple[tuple[int, ...], ...]:
        return tuple(itertools.product(*(
            range(len(subset)+1) for subset in self._EDGES.values()
        )))
    
    def subset_probability(self, weights, pi) -> Iterable[float]:
        raise NotImplementedError


from localuf._schemes import Forward
from localuf.decoders import UF

def test_get_syndrome(forward_rp: Forward):
    syndrome = forward_rp._get_syndrome(set(forward_rp._COMMIT_EDGES), set())
    assert syndrome == {v for v in forward_rp._CODE.NODES
                             if v[forward_rp._CODE.TIME_AXIS] == 0
                             and not forward_rp._CODE.is_boundary(v)}


def test_run(forward_rp: Forward):
    uf = UF(forward_rp._CODE)
    for n in range(1, 4):
        assert forward_rp.run(uf, 0, n) == (0, n+2)
        a, b = forward_rp.run(uf, 1, n)
        assert a != 0
        assert b == n+2


def test_make_error(forward3F: Forward):
    error = forward3F._make_error(set(), 0)
    assert error == set()
    error = forward3F._make_error(set(), 1)
    assert error == set(forward3F._BUFFER_EDGES)
    error = forward3F._make_error(set(forward3F._BUFFER_EDGES), 0)
    assert error == set(forward3F._COMMIT_EDGES)
    error = forward3F._make_error(set(forward3F._BUFFER_EDGES), 1)
    assert error == set(forward3F._CODE.EDGES)


"""Constants for local Union--Find decoder."""

from enum import IntEnum

BUSY_SIGNAL_SYMBOLS = {
    True: 'b',
    False: ' ',
}
"""How drawer should label a node with(out) a busy signal."""
ACTIVE_SIGNAL_SYMBOLS = {
    True: 'A',
    False: ' ',
}
"""How drawer should label a node with(out) an active signal."""


class Stage(IntEnum):
    """Constants for stage values."""

    INCREMENT = 1
    SV_STAGE_COUNT = 4
    """Syndrome validation stage count."""
    BP_STAGE_COUNT = 3
    """Burning & peeling stage count."""
    GROWING, MERGING, PRESYNCING, SYNCING = range(SV_STAGE_COUNT)
    BURNING, PEELING, DONE = range(SV_STAGE_COUNT, SV_STAGE_COUNT+BP_STAGE_COUNT)

    def __iadd__(self, other: 'Stage'):
        return Stage(self.value + other.value)

    def __imod__(self, other: 'Stage'):
        return Stage(self.value % other.value)

    def __str__(self):
        match self:
            case Stage.GROWING:    return 'G'
            case Stage.MERGING:    return 'M'
            case Stage.PRESYNCING: return 'PS'
            case Stage.SYNCING:    return 'S'
            case Stage.BURNING:    return 'B'
            case Stage.PEELING:    return 'P'
            case Stage.DONE:       return 'D'


import pytest

from localuf.decoders.snowflake import _Node, Friendship

@pytest.mark.parametrize("prop", [
    "NODE",
])
def test_property_attributes(test_property, sfn3: _Node, prop):
    test_property(sfn3.FRIENDSHIP, prop)


def test_drop(sfn3: _Node):
    fs = Friendship(sfn3)
    sfn3.unrooted = True
    sfn3.next_unrooted = True
    fs.drop()
    assert sfn3.unrooted is False
    assert sfn3.next_unrooted is False


"""Classes for noise models."""

import abc
from collections import defaultdict
from functools import cache
import random
from typing import Literal, Iterable

import numpy as np
import pandas as pd
from scipy.stats import binom

from localuf.noise.forcers import ForceByEdge, ForceByPair, ForceByPairBalanced
from localuf.type_aliases import Edge, EdgeType, Parametrization, FourFloats, MultiplicityVector, FourInts
from localuf.noise._multiset_handler import MultisetHandler


class Noise(abc.ABC):
    """Abstract base class for noise models."""

    @abc.abstractmethod
    def __str__(self) -> str: raise NotImplementedError

    @property
    @abc.abstractmethod
    def ALL_WEIGHTS(self) -> Iterable[int] | tuple[tuple[int, ...], ...]:
        """All possible ``force_error`` inputs."""
    
    @property
    @abc.abstractmethod
    def ALL_WEIGHTS_INDEX(self) -> pd.Index:
        """``ALL_WEIGHTS`` as pandas Index."""

    @abc.abstractmethod
    def make_error(self, noise_level: float) -> set[Edge]:
        """See ``Code.make_error``."""

    @abc.abstractmethod
    def force_error(self, weight: int | tuple[int, ...]) -> set[Edge]:
        """Make error of weight ``weight``."""
    
    @abc.abstractmethod
    def subset_probability(
        self,
        weights: Iterable[int] | tuple[tuple[int, ...], ...],
        noise_level: float,
    ) -> Iterable[float]:
        """Return probability of any error of weight ``weight``, for ``weight`` in ``weights``."""

    def subset_probabilities(self, noise_level: float, survival: bool = True):
        """Return DataFrame containing probabilities of each subset.
        
        
        :param noise_level: noise level.
        :param survival: whether to compute survival probability column.
        
        
        :returns: DataFrame indexed by subset weight, with columns ``['subset prob', 'survival prob']``.
        """
        subset_prob = self.subset_probability(self.ALL_WEIGHTS, noise_level)
        df = pd.DataFrame({'subset prob': subset_prob}, index=self.ALL_WEIGHTS_INDEX)
        if survival:
            df.sort_values(by='subset prob', inplace=True)
            df['survival prob'] = df['subset prob'].cumsum().shift(fill_value=0)
        df.sort_values(by='subset prob', ascending=False, inplace=True)
        return df

    @abc.abstractmethod
    def get_edge_weights(
        self,
        noise_level: None | float,
    ) -> dict[Edge, tuple[float, float]]:
        """Return map from edge to its flip probability and weight.
        
        
        :param noise_level: a probability that represents the noise strength.
            This defines the nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If not specified, all edges have flip probability 0 and weight 1.
        
        
        :return edge_weights: A map from each edge in the decoding window
            to the pair (flip probability, weight).
        """

    @staticmethod
    def log_odds_of_no_flip(flip_probability: float) -> float:
        """Convert flip probability to log odds of no flip."""
        return np.log10((1 - flip_probability) / flip_probability)


class _Uniform(Noise):
    """Base class for noise model where each edge has the same flip probability.
    
    Extends ``Noise``.
    """

    def __init__(self, fresh_edges: tuple[Edge, ...], all_edges: None | tuple[Edge, ...] = None):
        """
        :param fresh_edges: The edges of the freshly discovered region after a window raise
            if scheme is streaming; else, the edges of G.
        """
        self._FRESH_EDGES = fresh_edges
        self._ALL_EDGES = fresh_edges if all_edges is None else all_edges
        """Same as `self._EDGES` but for _all_ edges in the decoding window."""

    @property
    def FRESH_EDGES(self):
        """The edges of the freshly discovered region after a window raise
        if scheme is streaming; else, the edges of G.
        """
        return self._FRESH_EDGES

    def make_error(self, noise_level):
        return {e for e in self.FRESH_EDGES if random.random() < noise_level}
    
    def force_error(self, weight: int):
        """Bitlfip exactly ``weight`` edges in G.
        
        :param weight: Between 0 and ``len(self.EDGES)``.
        :return error: The set of bitflipped edges.
        """
        return set(random.sample(self.FRESH_EDGES, weight))

    @property
    def ALL_WEIGHTS(self) -> Iterable[int]:
        return range(len(self.FRESH_EDGES)+1)
    
    @property
    def ALL_WEIGHTS_INDEX(self):
        return pd.Index(self.ALL_WEIGHTS, name='weight')
    
    def subset_probability(self, weights: Iterable[int], noise_level: float) -> Iterable[float]:
        probs = binom.pmf(
            k=weights,
            n=len(self.FRESH_EDGES),
            p=noise_level,
        )
        return probs
    
    @cache
    def get_edge_weights(self, noise_level: None | float) -> dict[Edge, tuple[float, float]]:
        if noise_level is None:
            flip_probability = 0
            weight = 1
        else:
            flip_probability = noise_level
            weight = self.log_odds_of_no_flip(noise_level)
        return {e: (flip_probability, weight) for e in self._ALL_EDGES}


class CodeCapacity(_Uniform):
    """Code capacity noise model.
    
    Extends ``_Uniform``.
    """

    def __str__(self) -> str:
        return 'code capacity'


class Phenomenological(_Uniform):
    """Phenomenological noise model.
    
    Extends ``_Uniform``.
    """

    def __str__(self) -> str:
        return 'phenomenological'


class CircuitLevel(Noise):
    """Circuit-level depolarizing noise model."""

    _DEFAULT_MULTIPLICITIES: dict[EdgeType, FourInts] = {
        'S':               (4, 2, 1, 0),
        'E westmost':      (1, 0, 2, 0),
        'E bulk':          (2, 0, 1, 0),
        'E eastmost':      (1, 0, 1, 0),
        'U 3':             (3, 0, 1, 1),
        'U 4':             (4, 0, 0, 1),
        'SD':              (2, 0, 0, 0),
        'EU west corners': (2, 0, 1, 0),
        'EU east corners': (2, 0, 2, 0),
        'EU edge':         (3, 0, 1, 0),
        'EU centre':       (4, 0, 0, 0),
        'SEU':             (2, 0, 0, 0),
    }
    """A map from edge type to multiplicity vector e.g. ``{'S': (4, 2, 1, 0), ...}``."""

    _ALL_COEFFICIENTS: dict[Parametrization, FourFloats] = {
        'standard': (4/15, 8/15, 2/3, 1),
        'balanced': (4/15, 8/15, 8/15, 8/15),
        'ion trap': (4/15, 8/15, (2e-3)/3, 1e-2),
    }
    """A map from parametrization name to 4-vector c such that pi = c*p."""

    def __str__(self) -> str:
        return 'circuit-level'

    def __init__(
            self,
            fresh_edge_dict: dict[EdgeType, tuple[Edge, ...]],
            parametrization: Parametrization,
            demolition: bool,
            monolingual: bool,
            fresh_merges: dict[Edge, Edge] | None = None,
            force_by: Literal['pair', 'edge'] = 'pair',
            all_edge_dict: None | dict[EdgeType, tuple[Edge, ...]] = None,
            all_merges: None | dict[Edge, Edge] = None,
    ):
        """
        :param fresh_edge_dict: maps from edge type to tuple of edges.
            The union of all edge tuples in ``fresh_edge_dict`` is the freshly discovered region after a window raise
            if scheme is streaming; else, the edges of G.
        :param parametrization: name of parametrization.
        :param demolition: whether ancilla qubit measurement demolishes state.
        :param monolingual: whether measurements are native in only one basis.
        :param fresh_merges: maps from each redundant edge to its substitute.
        :param force_by: whether ``force_error()`` makes an error based on the weight of each pair or edge type.
            See ``noise.forcers`` for more details.
        """
        self._FRESH_EDGES = self._make_edges(fresh_edge_dict, demolition, monolingual, fresh_merges)
        """A map from multiplicity vector (as a tuple) to tuple of edges
        e.g. ``{(4, 2, 1, 0): (e1, ...), ...}``.
        Order matters as used by ``ForceByEdge.force_error``.
        The union of all edge tuples in ``_FRESH_EDGES`` is the freshly discovered region after a window raise
        if scheme is streaming; else, the edges of G.
        """
        self._COEFFICIENTS = self._ALL_COEFFICIENTS[parametrization]
        """A 4-vector c such that pi = c*p."""
        self._FORCER = ForceByEdge(self._FRESH_EDGES) if force_by == 'edge' \
            else ForceByPairBalanced(self._FRESH_EDGES) if parametrization == 'balanced' \
            else ForceByPair(self._FRESH_EDGES)
        """Method for forcing error."""
        self._ALL_EDGES = self._FRESH_EDGES if all_edge_dict is None else \
            self._make_edges(all_edge_dict, demolition, monolingual, all_merges)
        """Same as `self._FRESH_EDGES` but for _all_ edges in the decoding window."""

    @classmethod
    def _make_edges(
        cls,
        edge_dict: dict[EdgeType, tuple[Edge, ...]],
        demolition: bool,
        monolingual: bool,
        merges: dict[Edge, Edge] | None,
    ):
        """Return ``_EDGES`` instance attribute."""
        multiplicities = cls._make_multiplicities(demolition, monolingual)
        edges: defaultdict[FourInts, list[Edge]] = defaultdict(list)
        if merges is None:
            for edge_type, edges_of_that_type in edge_dict.items():
                multiplicity: FourInts = tuple(multiplicities[edge_type]) # type: ignore
                edges[multiplicity] += list(edges_of_that_type)
        else:
            dc = cls._get_dc(edge_dict, multiplicities, merges)
            for edge, multiplicity_vector in dc.items():
                multiplicity: FourInts = tuple(multiplicity_vector) # type: ignore
                edges[multiplicity].append(edge)
        return dict(edges)

    @classmethod
    def _make_multiplicities(cls, demolition: bool, monolingual: bool):
        """Return map from edge type to multiplicity vector.
        
        E.g. ``{'S': np.array((4, 2, 1, 0)), ...}``.
        """
        multiplicities: dict[EdgeType, MultiplicityVector] = {
            edge_type: np.array(m)
            for edge_type, m
            in cls._DEFAULT_MULTIPLICITIES.items()
        }
        horizontals: list[EdgeType] = [
            'S',
            'E westmost',
            'E bulk',
            'E eastmost',
        ]
        verticals: list[EdgeType] = ['U 3', 'U 4']
        if demolition:
            for edge_type in horizontals:
                multiplicities[edge_type][2] += 1
            for edge_type in verticals:
                multiplicities[edge_type][3] += 1
        if monolingual:
            for edge_type in horizontals+verticals:
                multiplicities[edge_type][2] += 2
        return multiplicities

    @staticmethod
    def _get_dc(
        edge_dict: dict[EdgeType, tuple[Edge, ...]],
        multiplicities: dict[EdgeType, MultiplicityVector],
        merges: dict[Edge, Edge],
    ):
        """Return map from edge to multiplicity 4-vector."""
        # IMPORTANT: instantiate NEW array for each edge!
        dc: defaultdict[Edge, MultiplicityVector] = defaultdict(lambda: np.zeros(4, dtype=int))
        for edge_type, edges in edge_dict.items():
            m = multiplicities[edge_type]
            for e in edges:
                if e in merges:
                    substitute = merges[e]
                    dc[substitute] += m
                else:
                    dc[e] += m
        return dc
    
    def make_error(self, noise_level):
        bitflip_probs = self._get_flip_probabilities(noise_level)
        error: set[Edge] = set()
        for m, pr in bitflip_probs.items():
            error |= {e for e in self._FRESH_EDGES[m] if random.random() < pr}
        return error
    
    def force_error(self, weight: tuple[int, ...]):
        return self._FORCER.force_error(weight)
    
    @property
    def ALL_WEIGHTS(self):
        return self._FORCER.ALL_WEIGHTS
    
    @property
    def ALL_WEIGHTS_INDEX(self):
        return pd.MultiIndex.from_tuples(self.ALL_WEIGHTS)
    
    def subset_probability(self, weights: tuple[tuple[int, ...], ...], noise_level: float):
        pi = self._pi(noise_level)
        return self._FORCER.subset_probability(weights, pi)
    
    @cache
    def get_edge_weights(self, noise_level: None | float):
        edge_weights: dict[Edge, tuple[float, float]] = {}
        if noise_level is None:
            for multiplicity_vector, edges in self._ALL_EDGES.items():
                for e in edges:
                    edge_weights[e] = (0, 1)
        else:
            flip_probabilities = self._get_flip_probabilities(noise_level)
            for multiplicity_vector, edges in self._ALL_EDGES.items():
                flip_probability = flip_probabilities[multiplicity_vector]
                weight = self.log_odds_of_no_flip(flip_probability)
                for e in edges:
                    edge_weights[e] = (flip_probability, weight)
        return edge_weights

    def _pi(self, noise_level: float) -> FourFloats:
        """Return 4-vector probability ``pi = c*noise_level``."""
        return tuple(c*noise_level for c in self._COEFFICIENTS) # type: ignore
    
    @cache
    def _get_flip_probabilities(self, noise_level: float) -> dict[FourInts, float]:
        """Return map from multiplicity to flip probability.
        
        
        :param noise_level: A probability that represents the noise strength.
        
        
        :return flip_probabilities: map from multiplicity to a flip probability.
        """
        pi = self._pi(noise_level)
        return {m: MultisetHandler.pr(m, pi) for m in self._FRESH_EDGES.keys()}



"""Module for local Union--Find decoder."""

from localuf.codes import Surface
from localuf._schemes import Batch, Frugal
from localuf.decoders.luf.constants import Stage
from localuf.decoders.luf.main import \
    LUF, Controller, Nodes, MacarNodes, ActisNodes, \
    Waiter, OptimalWaiter, UnoptimalWaiter, \
    _Node, MacarNode, ActisNode, \
    Friendship, ControllerFriendship, NodeFriendship


class Macar(LUF):
    """Local UF where controller directly connects to each node.
    
    Extends ``LUF``.
    Compatible only with surface code, and all schemes save frugal.
    """

    def __init__(self, code: Surface):
        """Input: ``code`` the code to be decoded."""
        if isinstance(code.SCHEME, Frugal):
            raise ValueError('Macar incompatible with frugal scheme.')
        super().__init__(code, visible=True)


class Actis(LUF):
    """Strictly local UF i.e. controller only connects to node whose ID is 0.
    
    Extends ``LUF``.
    Compatible only with surface code and batch scheme.
    """

    def __init__(self, code: Surface, _optimal=True):
        """
        :param code: the code to be decoded.
        :param _optimal: whether management of ``self.NODES.countdown`` is optimal.
        """
        if not isinstance(code.SCHEME, Batch):
            raise ValueError('Actis compatible only with batch scheme.')
        super().__init__(code, visible=False, _optimal=_optimal)


from localuf.decoders import Snowflake
from localuf._pairs import Pairs


def test_drop(snowflake3: Snowflake):
    pairs: Pairs = snowflake3.CODE.SCHEME.pairs # type: ignore

    u, v = (0, 0), (1, 0)
    edge = snowflake3.EDGES[u, v]

    edge.CONTACT.drop()
    assert pairs._dc == {}

    edge.correction = True
    edge.CONTACT.drop()
    assert edge.correction is True
    assert pairs._dc == {u: v, v: u}

    edge.CONTACT.drop()
    assert pairs._dc == {}


from unittest import mock

import pytest

from localuf.decoders import BUF
from localuf.decoders.uf import _Cluster

@pytest.fixture
def buf_after_union(
    buf: BUF,
    get_uf_after_union,
    syndrome7F,
) -> tuple[BUF, _Cluster, _Cluster]:
    buf.load(syndrome7F)
    return get_uf_after_union(buf)

def test_reset(buf: BUF):
    with mock.patch("localuf.decoders.uf.UF.reset") as super_reset:
        buf.reset()
        super_reset.assert_called_once_with()
    # test correct number of buckets
    assert len(buf.buckets) == buf.CODE.N_EDGES
    # test all buckets empty
    assert all(bucket == set() for bucket in buf.buckets.values())
    # test mvl
    assert buf.mvl is None

def test_load(buf: BUF, syndrome7F):
    with (
        mock.patch("localuf.decoders.uf.UF.load") as super_load,
        mock.patch("localuf.decoders.buf.BUF._update_mvl") as mock_update_mvl,
    ):
        buf.load(syndrome7F)
        super_load.assert_called_once_with(syndrome7F)
        mock_update_mvl.assert_called_once_with()
    # test buckets 3 & 4 initially have all defects
    assert {
        cluster.root for cluster in buf.buckets[3].union(buf.buckets[4])
    } == buf.syndrome
    # test active_clusters attribute replaced by buckets
    assert not hasattr(buf, 'active_clusters')

def test_validate(buf: BUF, syndrome7F):

    def set_mvl(mvl: None | int = None):
        buf.mvl = mvl

    not_none = 1
    set_mvl(not_none)
    with (
        mock.patch("localuf.decoders.buf.BUF.load") as mock_load,
        mock.patch("localuf.decoders.buf.BUF._growth_round") as mock_growth_round,
        mock.patch(
            "localuf.decoders.buf.BUF._update_mvl",
            side_effect=set_mvl,
        ) as mock_update_mvl,
    ):
        buf.validate(syndrome7F)
        mock_load.assert_called_once_with(syndrome7F)
        mock_growth_round.assert_called_once_with(
            False,
            clusters_to_grow=buf.buckets[not_none],
        )
        mock_update_mvl.assert_called_once_with()

def test_update_mvl(buf: BUF):
    buf.buckets[3] = {buf.clusters[0, 0]}
    buf._update_mvl()
    assert buf.mvl == 3

def test_update_self_after_union(
        buf_after_union,
        fixture_test_update_self_after_union,
):
    fixture_test_update_self_after_union(buf_after_union)


import numpy as np
import pytest

from localuf.type_aliases import Edge, MultiplicityVector
from localuf.noise import CircuitLevel


@pytest.fixture
def ottf():
    return (1, 2, 3, 4)


@pytest.fixture
def cl_dc(e_westmost: tuple[Edge, Edge], ottf):
    e0, e1 = e_westmost
    f = ((1, -1, 0), (1, 0, 1))
    g = ((0, -1, 0), (1, 0, 1))

    return CircuitLevel._get_dc(
        edge_dict={
            'E westmost': (e0, e1),
            'EU west corners': (f,),
            'SEU': (g,),
        },
        multiplicities={
            'E westmost': np.array(ottf),
            'EU west corners': 10*np.array(ottf),
            'SEU': 100*np.array(ottf),
        },
        merges={f: e1, g: e1},
    )


def test_get_dc(
        cl_dc: dict[Edge, MultiplicityVector],
        e_westmost: tuple[Edge, Edge],
        ottf,
):
    e0, e1 = e_westmost

    assert len(cl_dc) == 2
    assert (cl_dc[e0] == np.array(ottf)).all()
    assert (cl_dc[e1] == 111*np.array(ottf)).all()


def test_change_edges(e_westmost: tuple[Edge, Edge]):
    e0, e1 = e_westmost
    dc = CircuitLevel._get_dc(
        edge_dict={'E westmost': (e0, e1)},
        multiplicities={'E westmost': np.array((1, 2, 3, 4))},
        merges={},
    )
    assert (dc[e0] == dc[e1]).all()
    assert dc[e0] is not dc[e1]
    dc[e0] *= 2
    assert (dc[e0] == 2*dc[e1]).all()


"""Module for emulating runtime for Macar, Actis, Snowflake.

Available functions:

* batch
* frugal
"""

from collections import defaultdict
from collections.abc import Callable
import itertools
from typing import Iterable, Literal, Type

from pandas import DataFrame

from localuf import Surface
from localuf.decoders import Macar, Actis
from localuf.decoders.snowflake import Snowflake
from localuf.type_aliases import NoiseModel, StreamingNoiseModel
from localuf._base_classes import Code
from localuf._schemes import Forward, Frugal
from localuf.sim._height_calculator import get_heights


def batch(
        ds: Iterable[int],
        noise_levels: Iterable[float],
        n: int,
        noise: NoiseModel,
        decoder_class: Type[Macar | Actis] = Macar,
        validate_only=True,
        **kwargs_for_Surface,
):
    """Make runtime data of Local UF decoder under batch decoding scheme.
    
    
    :param ds: an iterable of surface code distances.
    :param noise_levels: an iterable of noise levels.
    :param n: sample count.
    :param noise: noise model.
    :param decoder_class: the class of the decoder.
    :param validate_only: whether to time only syndrome validation or the full decoding cycle.
    :param kwargs_for_Surface: passed to ``Surface``.
    
    :return data: A DataFrame where each
        column a (stage, distance, probability);
        row, a runtime sample.
        Stage is either ``SV`` or ``BP``
        which stand for 'syndrome validation' and 'burning & peeling'.
    """
    scheme = 'batch'
    dc: defaultdict[tuple[str, int, float], list[int]] = defaultdict(list)
    for d in ds:
        code = Surface(
            d,
            noise,
            scheme=scheme,
            window_height=None,
            **kwargs_for_Surface,
        )
        decoder = decoder_class(code)
        for noise_level in noise_levels:
            for _ in itertools.repeat(None, times=n):
                error = code.make_error(noise_level)
                syndrome = code.get_syndrome(error)
                if validate_only:
                    tSV = decoder.validate(syndrome)
                else:
                    tSV, tBP = decoder.decode(syndrome)
                    dc['BP', d, noise_level].append(tBP)
                dc['SV', d, noise_level].append(tSV)
                decoder.reset()
    data = DataFrame(dc).sort_index(axis=1)
    data.columns.set_names(['stage', 'd', 'p'], inplace=True)
    return data


def forward(
        ds: Iterable[int],
        noise_levels: Iterable[float],
        n: int,
        noise: StreamingNoiseModel,
        get_commit_height: Callable[[int], int] | None = None,
        get_buffer_height: Callable[[int], int] | None = None,
        **kwargs_for_Surface,
):
    """Make runtime data of Macar under forward decoding scheme.
    
    Use this function to analyse throughput.
    Latency can be analysed using ``sim.runtime.batch``,
    since the final decoding window of a memory experiment
    resembles that of a batch decoding window.
    
    
    :param ds: same as for ``batch``.
    :param noise_levels: same as for ``batch``.
    :param n: same as for ``batch``.
    :param noise: same as for ``batch``.
    :param kwargs_for_Surface: same as for ``batch``.
    :param get_commit_height: a function with input ``d`` that outputs commit height
        e.g. ``lambda d: 2*(d//2)``.
        If ``None``, commit height is ``d``.
    :param get_buffer_height: a function with input ``d`` that outputs buffer height.
        If ``None``, buffer height is ``d``.
    
    :return: Same as for ``batch`` where ``validate_only=True``.
    """
    scheme = 'forward'
    dc: dict[tuple[str, int, float], list[int]] = {}
    for d in ds:
        _, commit_height, buffer_height = get_heights(
            d,
            n,
            scheme,
            get_commit_height=get_commit_height,
            get_buffer_height=get_buffer_height,
        )
        code = Surface(
            d,
            noise,
            scheme=scheme,
            window_height=None,
            commit_height=commit_height,
            buffer_height=buffer_height,
            **kwargs_for_Surface,
        )
        decoder = Macar(code)
        for noise_level in noise_levels:
            forward: Forward = code.SCHEME # type: ignore
            forward.run(decoder, noise_level, n)
            step_counts: list[tuple[int, int]] = forward.step_counts # type: ignore
            dc['SV', d, noise_level] = [sv for sv, _ in step_counts]
            dc['BP', d, noise_level] = [bp for _, bp in step_counts]
    data = DataFrame(dc).sort_index(axis=1)
    data.columns.set_names(['stage', 'd', 'p'], inplace=True)
    return data


def frugal(
        ds: Iterable[int],
        noise_levels: Iterable[float],
        n: int,
        code_class: Type[Code],
        noise: StreamingNoiseModel,
        time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
        get_commit_height: Callable[[int], int] | None = None,
        get_buffer_height: Callable[[int], int] | None = None,
        **kwargs_for_Snowflake,
):
    """Make runtime data Snowflake.
    
    
    :param ds: an iterable of surface code distances.
    :param noise_levels: an iterable of noise levels.
    :param n: sample count.
    :param code_class: the class of the code.
    :param noise: the noise model.
    :param time_only: whether runtime includes a timestep
        for each drop, each grow, and each merging step ('all');
        each merging step only ('merging');
        or each unrooting step only ('unrooting').
    :param get_commit_height: a function with input ``d`` that outputs commit height.
        If ``None``, commit height is ``1``.
    :param get_buffer_height: a function with input ``d`` that outputs buffer height.
        If ``None``, buffer height is ``2*(d//2)``.
    :param kwargs_for_Snowflake: passed to Snowflake
        e.g. ``merger`` decides whether Snowflake's nodes
        flood before syncing (fast) or vice versa (slow) in a merging step.
    
    Output: ``data`` a DataFrame where each
    column a (distance, probability);
    row, a runtime sample.
    """
    scheme = 'frugal'
    dc: dict[tuple[int, float], tuple[int, ...]] = {}
    for d in ds:
        _, commit_height, buffer_height = get_heights(
            d,
            n,
            scheme,
            get_commit_height=get_commit_height,
            get_buffer_height=get_buffer_height,
        )
        code = code_class(
            d,
            noise,
            scheme=scheme,
            window_height=None,
            commit_height=commit_height,
            buffer_height=buffer_height,
        )
        frugal: Frugal = code.SCHEME # type: ignore
        decoder = Snowflake(code, **kwargs_for_Snowflake)
        for noise_level in noise_levels:
            frugal.run(decoder, noise_level, n, time_only=time_only)
            dc[d, noise_level] = tuple(frugal.step_counts)
    data = DataFrame(dc).sort_index(axis=1)
    data.columns.set_names(['d', 'p'], inplace=True)
    return data


from math import prod

from localuf.type_aliases import FourFloats, FourInts, MultiplicityVector


class MultisetHandler:

    @staticmethod
    def pr(multiplicities: FourInts | MultiplicityVector, pi: FourFloats) -> float:
        """Return probability of odd number of faults in multiset
        defined by ``multiplicities``, ``pi``.
        
        
        :param multiplicities: a multiplicity vector.
        :param pi: a 4-tuple of fault probabilities.
        """
        f1 = prod((1-p)**m for p, m in zip(pi, multiplicities))
        f2 = sum(m*p/(1-p) for p, m in zip(pi, multiplicities))
        return f1 * f2


from localuf.type_aliases import Node


class Determinant:
    """Base class for determining whether a node is a boundary.
    
    Instance attributes:
    
    * ``D`` code distance.
    * ``LONG_AXIS`` that whose index runs from -1 to d-1 inclusive.
    """

    def __init__(self, d: int, long_axis: int) -> None:
        self._D = d
        self._LONG_AXIS = long_axis

    @property
    def D(self): return self._D

    @property
    def LONG_AXIS(self): return self._LONG_AXIS

    def is_boundary(self, v: Node):
        """See ``_base_classes.Code.is_boundary``."""
        # `node` either (i, j) or (j, t) or (i, j, t).
        return v[self.LONG_AXIS] in {-1, self.D-1}


class SpaceDeterminant(Determinant):
    """Determines only space boundaries.
    
    Extends ``Determinant``.
    """


class SpaceTimeDeterminant(Determinant):
    """Determines space and time boundaries.
    
    Extends ``Determinant``.
    
    Additional instance attributes:
    
    * ``TIME_AXIS`` that which represents time.
    * ``WINDOW_HEIGHT`` total height of sliding window.
    
    Overriden methods:
    
    * ``__init__``
    * ``is_boundary``
    """

    def __init__(
            self,
            d: int,
            long_axis: int,
            time_axis: int,
            window_height: int
    ):
        super().__init__(d, long_axis)
        self._TIME_AXIS = time_axis
        self._WINDOW_HEIGHT = window_height

    @property
    def TIME_AXIS(self): return self._TIME_AXIS
    
    @property
    def WINDOW_HEIGHT(self): return self._WINDOW_HEIGHT

    def is_boundary(self, v: Node):
        is_space_boundary = super().is_boundary(v)
        is_time_boundary = v[self.TIME_AXIS]==self.WINDOW_HEIGHT
        return is_space_boundary or is_time_boundary


import itertools

import pytest

from localuf import Repetition, Surface
from localuf.type_aliases import Node, Edge
from localuf.constants import Growth
from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node

whole = True

@pytest.fixture
def snowflake3tall():
    d = 3
    rp = Repetition(
        d,
        noise='phenomenological',
        scheme='frugal',
        buffer_height=3*(d//2),
    )
    return Snowflake(rp)


@pytest.fixture
def snowflake4():
    d = 4
    rp = Repetition(
        d,
        noise='phenomenological',
        scheme='frugal',
    )
    return Snowflake(rp, merger='slow', schedule='1:1')


@pytest.fixture
def snowflake5():
    d = 5
    rp = Repetition(
        d,
        noise='phenomenological',
        scheme='frugal',
    )
    return Snowflake(rp, schedule='1:1')


@pytest.fixture
def snowflake7():
    d = 7
    rp = Repetition(
        d,
        noise='phenomenological',
        scheme='frugal',
        buffer_height=3*(d//2),
    )
    return Snowflake(rp, merger='slow', schedule='1:1')


def test_nonroot_boundary_defect_kept(snowflake4: Snowflake):
    """Ensure a nonroot boundary node w/ defect loses not its defect."""
    h = snowflake4.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in [
        {(0, h-1), (2, h-1)},
        set(),
        {(0, h-1)},
        set(),
    ]:
        snowflake4.decode(syndrome)
    snowflake4.drop(set())
    snowflake4._SCHEDULE.grow()
    for _ in itertools.repeat(None, 5):
        for node in snowflake4.NODES.values():
            node.merging(whole)
        for node in snowflake4.NODES.values():
            node.update_after_merging()
        assert snowflake4.NODES[-1, 0].defect


def assert_no_standoff(s: Snowflake):
    """Assert no pointer standoff in the snowflake."""
    for u, v in s.EDGES.keys():
        node_u = s.NODES[u]
        node_v = s.NODES[v]
        assert_no_standoff_between(node_u, node_v)

def assert_no_standoff_between(node_u: _Node, node_v: _Node):
    """Assert no pointer standoff between `node_u` and `node_v`."""
    if node_u.pointer != 'C' and node_v.pointer != 'C':
        assert not all((
                node_u.access[node_u.pointer] == node_v,
                node_v.access[node_v.pointer] == node_u,
            ))

def test_decode_coarse(snowflake3tall: Snowflake):
    """Test adjacent pointers never point toward each other
    after each decode.
    """
    h = snowflake3tall.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in itertools.chain(
        itertools.repeat({(1, h-1)}, 2),
        itertools.repeat({(0, h-1), (1, h-1)}, 1),
        itertools.repeat(set(), 2*h-1),
    ):
        snowflake3tall.decode(syndrome)
        assert_no_standoff(snowflake3tall)


def test_decode_fine(snowflake5: Snowflake):
    """Test adjacent pointers never point toward each other
    after each merging timestep.
    """
    h = snowflake5.CODE.SCHEME.WINDOW_HEIGHT
    snowflake5.decode({(0, h-1)})
    snowflake5.decode({(1, h-1)})
    snowflake5.decode({(0, h-1)})
    for _ in itertools.repeat(None, 2*h-1):
        # DECODE
        snowflake5.drop(set())
        snowflake5._SCHEDULE.grow()
        while True:
            for node in snowflake5.NODES.values():
                node.merging(whole)
            for node in snowflake5.NODES.values():
                node.update_after_merging()
            assert_no_standoff(snowflake5)
            if not any(node.busy for node in snowflake5.NODES.values()):
                break


def test_snake(snowflake5: Snowflake):
    """Test scenario which 'keep & peel' method fails.
    
    This is because the ((0, 0), (1, 0)) edge incorrectly points west after first drop.
    """
    for e in {
        ((0, 1), (1, 1)),
        ((1, 1), (2, 1)),
        ((0, 2), (1, 2)),
        ((1, 2), (1, 3)),
        ((0, 3), (1, 3)),
        ((-1, 3), (0, 3)),
        ((0, 1), (0, 2)),
        ((2, 1), (2, 2)),
    }:
        snowflake5.EDGES[e].growth = Growth.FULL
    for _ in itertools.repeat(None, 4):
        snowflake5.decode(set())
        assert_no_standoff(snowflake5)


def test_no_lost_defect(snowflake5: Snowflake):
    """Test no defect in bottom layer of detectors after each growth round.
    
    This uses a J-shaped cluster whose root is a boundary node
    at the bottom end of the J.

    The method that fails this test is the one that on window raise preserves...
    * CID structure unless CID is one of the lowest boundary nodes
    (in which case CID resets)
    * pointer structure unless pointer points down to nothing
    (in which case pointer resets)
    """
    h = snowflake5.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in itertools.chain(
        [
            {(2, h-1), (0, h-1)},
            {(2, h-1)},
            {(2, h-1), (0, h-1)},
        ],
        itertools.repeat({(2, h-1)}, 2),
        itertools.repeat(set(), 2*h),
    ):
        snowflake5.decode(syndrome)
        for j in range(snowflake5.CODE.D-1):
            assert not snowflake5.NODES[j, 0].defect


def assert_lowest_nodes_never_point_down(decoder: Snowflake, syndrome: set[Node]):
    """Call `decoder.decode(syndrome)` then assert lowest nodes never point down."""
    decoder.decode(syndrome)
    for j in range(decoder.CODE.D-1):
        assert decoder.NODES[j, 0].pointer != 'D'


def test_no_pointer_standoff_after_keep(snowflake5: Snowflake):
    """Test no pointer standoff after a cluster merges with a kept cluster (legacy)."""
    h = snowflake5.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in itertools.chain(
        [
            {(3, h-1)},
            {(2, h-1)},
            {(1, h-1)},
            {(3, h-1)},
        ],
        itertools.repeat(set(), 2*h-1),
    ):
        assert_lowest_nodes_never_point_down(snowflake5, syndrome)


def test_detached_cluster_keep(snowflake5: Snowflake):
    """Test unpeeled nodes have `keep` True when in a detached cluster (legacy).
    
    This uses a U-shaped cluster with no boundary node,
    whose root is at the top left end of the U.

    This test fails if the bottom of the cluster disappears
    before all of the cluster is under the floor.
    """
    h = snowflake5.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in itertools.chain(
        [
            {(0, h-1), (1, h-1), (2, h-1), (3, h-1)},
            {(0, h-1), (3, h-1)},
        ],
        itertools.repeat(set(), 2*h-1),
    ):
        assert_lowest_nodes_never_point_down(snowflake5, syndrome)


def helper_infinite_due_to_unroot(snowflake5: Snowflake):
    h = snowflake5.CODE.SCHEME.WINDOW_HEIGHT
    for syndrome in [
        {(2, h-1),},
        {(2, h-1), (3, h-1),},
        {(1, h-1),},
        {(2, h-1),},
        set(),
        set(),
    ]:
        snowflake5.decode(syndrome)
    snowflake5.drop(set())
    snowflake5._SCHEDULE.grow()

def test_no_infinite_unroot_cycle(snowflake5: Snowflake):
    """Test unroot wave never cycles infinitely.
    
    This fails if:
    * unroot wave propagates along pointers rather than access,
    * and `unrooted` is not used.
    """
    helper_infinite_due_to_unroot(snowflake5)
    for _ in itertools.repeat(None, 10):  # prevent infinite loop
        for node in snowflake5.NODES.values():
            node.merging(whole)
        for node in snowflake5.NODES.values():
            node.update_after_merging()
        if not any(node.busy for node in snowflake5.NODES.values()):
            break
    for v in [
        (0, 0),
        (1, 0),
        (1, 1),
        (2, 0),
        (2, 1),
        (2, 2),
        (3, 1),
    ]:
        assert snowflake5.NODES[v].cid == 20

def test_no_unroot_pointer_cycle(snowflake5: Snowflake):
    """Test unroot never leads to pointer cycle.
    
    This fails if unroot wave propagates along pointers rather than access.
    """
    helper_infinite_due_to_unroot(snowflake5)
    for _ in itertools.repeat(None, 10):  # prevent infinite loop
        for node in snowflake5.NODES.values():
            node.merging(whole)
        for node in snowflake5.NODES.values():
            node.update_after_merging()
        assert not all((
            snowflake5.NODES[1, 0].pointer == 'E',
            snowflake5.NODES[2, 0].pointer == 'U',
            snowflake5.NODES[2, 1].pointer == 'W',
            snowflake5.NODES[1, 1].pointer == 'D',
        ))
        if not any(node.busy for node in snowflake5.NODES.values()):
            break


def test_no_infinite_loop(snowflake7: Snowflake):
    """Test infinite loop scenario avoided.
    
    The infinite loop tested is a defect bouncing between 2 nodes forever.
    """
    h = snowflake7.CODE.SCHEME.WINDOW_HEIGHT
    snowflake7.decode({(0, h-1), (1, h-1)})
    for _ in itertools.repeat(None, 6):
        snowflake7.decode({(1, h-1)})
    snowflake7.decode({(1, h-1), (3, h-1)})
    snowflake7.decode({(0, h-1), (1, h-1)})
    snowflake7.decode(set())

    # DECODE
    snowflake7.drop(set())
    snowflake7._SCHEDULE.grow()
    for cycle_index in range(1, 11):
        for node in snowflake7.NODES.values():
            node.merging(whole)
        for node in snowflake7.NODES.values():
            node.update_after_merging()
        if cycle_index == 11:
            for v, node in snowflake7.NODES.items():
                # check there's 1 defect left
                if not snowflake7.CODE.is_boundary(v):
                    assert node.defect is (v==(1, 7))
                # check that defect won't bounce between 2 nodes forever
                node_u = snowflake7.NODES[0, 7]
                node_v = snowflake7.NODES[1, 7]
                assert_no_standoff_between(node_u, node_v)
        if not any(node.busy for node in snowflake7.NODES.values()):
            break


class TestLowestEdges:

    @staticmethod
    def _check_edge_indices(decoder: Snowflake, indices: tuple[Edge, ...]):
        for index, edge in zip(indices, decoder._LOWEST_EDGES, strict=True):
            assert edge.INDEX == index


    def test_repetition(self, snowflake3: Snowflake):
        indices = (
            ((0, 0), (0, 1)),
            ((1, 0), (1, 1)),
            ((-1, 0), (0, 0)),
            ((0, 0), (1, 0)),
            ((1, 0), (2, 0)),
        )
        self._check_edge_indices(snowflake3, indices)


    def test_surface(self, surface3_CL_frugal: Surface):
        """Verified `indices` by drawing each edge in a notebook."""
        decoder = Snowflake(surface3_CL_frugal)
        indices = (
            ((0, 0, 0), (0, 0, 1)),
            ((0, 1, 0), (0, 1, 1)),
            ((1, 0, 0), (1, 0, 1)),
            ((1, 1, 0), (1, 1, 1)),
            ((2, 0, 0), (2, 0, 1)),
            ((2, 1, 0), (2, 1, 1)),
            ((0, 0, 1), (1, 0, 0)),
            ((0, 1, 1), (1, 1, 0)),
            ((1, 0, 1), (2, 0, 0)),
            ((1, 1, 1), (2, 1, 0)),
            ((0, 0, 0), (0, 1, 1)),
            ((1, 0, 0), (1, 1, 1)),
            ((2, 0, 0), (2, 1, 1)),
            ((0, 0, 0), (1, 1, 1)),
            ((1, 0, 0), (2, 1, 1)),
            ((0, 0, 0), (1, 0, 0)),
            ((0, 1, 0), (1, 1, 0)),
            ((1, 0, 0), (2, 0, 0)),
            ((1, 1, 0), (2, 1, 0)),
            ((0, -1, 0), (0, 0, 0)),
            ((0, 0, 0), (0, 1, 0)),
            ((0, 1, 0), (0, 2, 0)),
            ((1, -1, 0), (1, 0, 0)),
            ((1, 0, 0), (1, 1, 0)),
            ((1, 1, 0), (1, 2, 0)),
            ((2, -1, 0), (2, 0, 0)),
            ((2, 0, 0), (2, 1, 0)),
            ((2, 1, 0), (2, 2, 0)),
        )
        self._check_edge_indices(decoder, indices)


from unittest import mock

import pytest

from localuf.constants import Growth
from localuf.decoders.snowflake import Snowflake, _Edge, FloorContact, EdgeContact

@pytest.mark.parametrize("prop", [
    "SNOWFLAKE",
    "INDEX",
    "CONTACT",
])
def test_property_attributes(test_property, sfe3: _Edge, prop):
    test_property(sfe3, prop)


def test_init(sfe3: _Edge):
    with mock.patch("localuf.decoders.snowflake._Edge.reset") as mock_reset:
        sfe3.__init__(sfe3.SNOWFLAKE, sfe3.INDEX)
        mock_reset.assert_called_once_with()


def test_SNOWFLAKE_property(sfe3: _Edge):
    assert type(sfe3.SNOWFLAKE) is Snowflake
    assert sfe3.SNOWFLAKE.EDGES[sfe3.INDEX] == sfe3


def test_INDEX_property(sfe3: _Edge, fixture_test_INDEX_property):
    fixture_test_INDEX_property(sfe3)


def test_CONTACT_property(snowflake: Snowflake):
    for (u, v), edge in snowflake.EDGES.items():
        ut = u[snowflake.CODE.TIME_AXIS]
        vt = v[snowflake.CODE.TIME_AXIS]
        if ut != 0 and vt != 0:
            assert type(edge.CONTACT) is EdgeContact
        else:
            assert type(edge.CONTACT) is FloorContact


def test_reset(sfe3: _Edge):

    sfe3.growth = Growth.FULL
    sfe3.correction = True

    sfe3.next_growth = Growth.FULL
    sfe3.next_correction = True

    sfe3.reset()

    assert sfe3.growth is Growth.UNGROWN
    assert sfe3.correction is False

    assert sfe3.next_growth is Growth.UNGROWN
    assert sfe3.next_correction is False


def test_update_after_drop(sfe3: _Edge):
    sfe3.next_growth = Growth.FULL
    sfe3.next_correction = True
    sfe3.update_after_drop()
    assert sfe3.growth is Growth.FULL
    assert sfe3.correction is True
    assert sfe3.next_growth is Growth.FULL
    assert sfe3.next_correction is True


from localuf import Surface

def test_init():
    """Calls `Surface._substitute`.
    
    Fails if `Surface._substitute` calls `self._NOISE` before this is defined.
    """
    code = Surface(3, 'circuit-level', merge_equivalent_boundary_nodes=True)
    assert code.DIMENSION == 3


from abc import abstractmethod
import copy
from functools import cache, cached_property
from typing import Literal

import networkx as nx

from localuf import constants
from localuf._base_classes import Code, Decoder
from localuf.constants import Growth
from localuf.type_aliases import Edge, Node
from itertools import chain, repeat


direction = Literal[
    'C',
    'N',
    'W',
    'E',
    'S',
    'D',
    'U',
    'SD',
    'NU',
    'EU',
    'WD',
    'SEU',
    'NWD',
]
"""Possible pointer value."""


class BaseUF(Decoder):
    """The abstract class representing the UF decoder (Union--Find).
    
    Extends ``BaseDecoder``.
    """

    _ACTIVE_GROWTH_VALUES = {Growth.UNGROWN, Growth.HALF}
    """The set of growth values for which an edge is active i.e. can grow."""
    _INACTIVE_GROWTH_VALUES = {Growth.BURNT, Growth.FULL}
    """The set of growth values for which an edge is inactive."""

    def __init__(self, code: Code):
        """Input: ``code`` the code to be decoded."""
        super().__init__(code)
        self._growth: dict[Edge, Growth] = {e: Growth.UNGROWN for e in self.CODE.EDGES}
        self.syndrome: set[Node]
        """The set of defects."""
        self.history: list[BaseUF]
        """A list of past self snapshots at each growth round
        (all with the same ``history`` attribute to prevent infinite loop).
        """

    @property
    def growth(self):
        """A map from each edge in the decoding graph to an integer representing its growth value."""
        return self._growth

    @property
    @abstractmethod
    def _FIG_WIDTH(self) -> float:
        """Width of figure for drawer."""
    
    @property
    @abstractmethod
    def _FIG_HEIGHT(self) -> float:
        """Height of figure for drawer."""

    @cached_property
    def erasure(self):
        """The set of fully grown edges, computed from ``self.growth``.
        Only computed if ``erasure`` not yet an attribute.
        """
        return {e for e, growth in self.growth.items() if growth is Growth.FULL}

    def reset(self):
        """Factory reset."""
        # use `_growth` instead of `growth` as the child class `Snowflake`
        # defines latter as a property that must be computed every time it is accessed
        # meaning the property would have to be redundantly computed |E| times!
        for e in self._growth.keys():
            self._growth[e] = Growth.UNGROWN
        super().reset()
        try: del self.erasure
        except AttributeError: pass
        try: del self.history
        except AttributeError: pass

    def init_history(self):
        """Initialize ``history`` attribute."""
        self.history = []

    def append_history(self):
        """Append a snapshot of the current ``self`` to ``self.history``."""
        self.history.append(copy.deepcopy(
            self, memo={id(self.history): self.history}
        ))

    @cache
    def _total_edge_weight(self, noise_level: None | float = None) -> float:
        """Compute the sum of all edge weights in the decoding window."""
        edge_weights = self.CODE.NOISE.get_edge_weights(noise_level)
        return sum(edge_weights[edge][1] for edge in self.growth.keys())
    
    def unclustered_edge_fraction(self, noise_level: None | float = None):
        """Compute the Unclustered Edge Fraction DCS from ``self.growth``.
        
        
        :param noise_level: A probability representing the noise strength.
            This is needed to define nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If not specified, all edges are assumed to have weight 1.
        
        
        :returns fraction: The fraction of edges in the decoding window
            that are not in a cluster, weighted by their weights.
        """
        edge_weights = self.CODE.NOISE.get_edge_weights(noise_level)
        numerator = sum(
            edge_weights[edge][1] * growth_value.as_float
            for edge, growth_value in self.growth.items()
        )
        return 1 - numerator / self._total_edge_weight(noise_level=noise_level)
    
    def draw_growth(
        self,
        growth: dict[Edge, Growth] | None = None,
        syndrome: set[Node] | None = None,
        highlighted_edges: set[Edge] | None = None,
        highlighted_edge_color='k',
        unhighlighted_edge_color=constants.GRAY,
        x_offset=constants.DEFAULT_X_OFFSET,
        with_labels=True,
        labels: dict[Node, str] | None = None,
        node_size=constants.SMALL,
        width=constants.WIDE_MEDIUM,
        boundary_color=constants.BLUE,
        defect_color=constants.RED,
        nondefect_color=constants.GREEN,
        show_boundary_defects=True,
        **kwargs_for_networkx_draw,
    ):
        """Draw growth of edges using matplotlib.
        
        
        :param growth: A dictionary where each key an edge index; value, its growth value.
        :param syndrome: The set of defects.
        :param highlighted_edges: The set of edges to be highlighted in drawing.
        :param x_offset: The ratio of out-of-screen to along-screen distance.
        :param with_labels: Whether to show node labels.
        :param labels: A dictionary where each key a node index; value, its label as a string.
        :param width: Line width of edges.
        :param arrows: Whether to draw arrows on edges used by pointers.
            For ``Macar, Actis, Snowflake`` decoders only.
        :param kwargs_for_networkx_draw: Passed to ``networkx.draw()``.
            E.g. ``linewidths`` line width of node symbol border.
        """
        g = self.CODE.GRAPH
        pos = self.CODE.get_pos(x_offset)
        if growth is None:
            growth = self.growth
        if syndrome is None:
            syndrome = set()
        if highlighted_edges is None:
            highlighted_edges = set()

        # node-related kwargs
        if labels is None:
            with_labels = False
        node_color = self.CODE.get_node_color(
            syndrome,
            boundary_color=boundary_color,
            defect_color=defect_color,
            nondefect_color=nondefect_color,
            show_boundary_defects=show_boundary_defects,
        )

        # edge-related kwargs
        edgelist = [
            e for e in self.CODE.EDGES
            if growth[e] in {Growth.HALF, Growth.FULL}
        ]
        edge_color, style = self._get_edge_color_and_style(
            edgelist,
            highlighted_edges,
            highlighted_edge_color,
            unhighlighted_edge_color,
            growth=growth,
        )
        return nx.draw(
            g,
            pos=pos,
            with_labels=with_labels,
            nodelist=self.CODE.NODES,
            edgelist=edgelist,
            node_size=node_size,
            node_color=node_color,
            width=width,
            edge_color=edge_color,
            style=style,
            labels=labels,
            **kwargs_for_networkx_draw,
        )

    def _get_edge_color_and_style(
            self,
            edgelist: list[Edge],
            highlighted_edges: set[Edge],
            highlighted_edge_color: str,
            unhighlighted_edge_color: str,
            growth: dict[Edge, Growth] | None = None,
    ):
        """Return ``edge_color``, ``style`` kwargs for ``networkx.draw()``."""
        if growth is None:
            growth = self.growth
        edge_color = [
            highlighted_edge_color if e in highlighted_edges else
            unhighlighted_edge_color for e in edgelist
        ]
        style = [
            ':' if growth[e] is Growth.HALF else
            '-' for e in edgelist
        ]
        return edge_color, style

    def _get_node_color_and_edgecolors(
            self,
            outlined_nodes: set[Node],
            nodelist: list[Node],
            outline_color: str = 'k',
            **kwargs_for_get_node_color,
    ):
        """Return node color and node edge color kwargs for ``networkx.draw()``.
        
        
        :param outlined_nodes: Set of nodes to be outlined.
        :param nodelist: List of nodes (needed to specify the node order in the outputs).
        :param kwargs_for_get_node_color: Passed to ``self.CODE.get_node_color()``.
        
        :return node_color: List of node colors.
        :return edgecolors: List of node edge colors.
        """
        node_color = self.CODE.get_node_color(
            self.syndrome,
            nodelist=nodelist,
            **kwargs_for_get_node_color,
        )
        edgecolors = [outline_color if v in outlined_nodes else color
                        for v, color in zip(nodelist, node_color)]
        return node_color, edgecolors
    
    def swim_distance(
            self,
            noise_level: None | float = None,
            draw=False,
            **kwargs_for_draw_swim_graph,
    ) -> float:
        """Compute the swim distance DCS from ``self.growth``.
        
        Assumes ``self.validate()`` or ``self.decode()`` has already been called.
        swim distance is the shortest distance from one boundary to the other
        where travelling within clusters is free.
        Introduced in [doi.org/10.1038/s42005-024-01883-4] and [arXiv:2405.07433].
        
        This method works regardless of whether ``self.CODE`` was instantiated with
        ``merge_equivalent_boundary_nodes=False`` or ``True``
        but works slightly faster in the latter case.
        
        :param noise_level: An optional probability representing the noise strength.
            This defines nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If not specified, all edges are assumed to have weight 1.
        :param draw: Whether to draw the search graph and the shortest path.
        :param kwargs_for_draw_swim_graph: Passed to ``self._draw_swim_graph()``.
        
        :return length: The swim distance.
        """
        graph = self._swim_graph(noise_level=noise_level)
        length, path = nx.bidirectional_dijkstra(graph, *self._META_BOUNDARIES)
        if draw:
            self._draw_swim_graph(path, **kwargs_for_draw_swim_graph)
        return length
    
    def _swim_graph(self, noise_level: None | float = None):
        """Return the search graph and endpoints for the swim distance.
        
        :param noise_level: As described in ``swim_distance()``.
        
        :return graph: A NetworkX graph representing
            the search graph for the swim distance DCS.
            The edges not in ``self.growth`` have weight 0 ALWAYS.
            The weight of the other edges is set according to ``self.growth``.
        """
        # TODO: use `networkx.quotient_graph`
        graph = self._cached_swim_graph
        for e, (_, weight) in self.CODE.NOISE.get_edge_weights(noise_level).items():
            if e not in self.growth or self.growth[e] is Growth.UNGROWN:
                # this can happen when the decoder omits some edges in W
                # e.g. Snowflake omitting the future boundary
                graph.edges[e]['weight'] = weight
            elif self.growth[e] is Growth.HALF:
                graph.edges[e]['weight'] = weight / 2
            else:  # growth is Growth.FULL or Growth.BURNT
                graph.edges[e]['weight'] = 0
        return self._cached_swim_graph
    
    @cached_property
    def _META_BOUNDARIES(self) -> tuple[Node, Node]:
        """The two extra nodes connecting the west and east boundary nodes.
        
        :return west: The node connected to the west boundary nodes.
        :return east: The node connected to the east boundary nodes.
        """
        d = self.CODE.D
        a = self.CODE.LONG_AXIS
        n = self.CODE.DIMENSION
        return (
            tuple(chain(repeat(d//2, a), (-2,), repeat(d//2, n-a-1))),
            tuple(chain(repeat(d//2, a), ( d,), repeat(d//2, n-a-1))),
        )
    
    @cached_property
    def _cached_swim_graph(self):
        """Return the search graph and endpoints for the swim distance.
        
        :return graph: A NetworkX graph representing
            the search graph for the swim distance DCS.
            The only time it is retrieved is in the property ``_swim_graph``,
            which always updates the edge weights before returning the graph.
            The edges not in ``self.growth`` have weight 0 ALWAYS.
            There is no guarantee about the weight of the other edges.
        """
        graph = self.CODE.GRAPH.copy()
        d = self.CODE.D
        a = self.CODE.LONG_AXIS
        west, east = self._META_BOUNDARIES
        graph.add_edges_from(((west, v) for v in self.CODE.NODES if v[a]==-1), weight=0)
        graph.add_edges_from(((v, east) for v in self.CODE.NODES if v[a]==d-1), weight=0)
        return graph

    def _draw_swim_graph(
            self,
            path: list[Node],
            weightless_edge_width=constants.THIN,
            max_edge_width=constants.V_WIDE,
            node_size=100,
            node_alpha=1,
            edge_alpha=0.5,
            **kwargs_for_networkx_draw,
    ):
        """Draw the search graph for the swim distance.
        
        :param path: The shortest west--east path.
        :param weightless_edge_width: The width of zero-weight edges.
        :param max_edge_width: The maximum width of edges.
        :param kwargs_for_networkx_draw: Passed to ``networkx.draw()``.
        """
        graph = self._cached_swim_graph
        max_weight: float = max(weight for _, _, weight in graph.edges.data('weight')) # type: ignore
        width_multiplier = (max_edge_width - weightless_edge_width) / max_weight
        nx.draw_networkx_nodes(
            graph,
            pos=self.CODE.get_pos(nodelist=graph.nodes),
            node_size=node_size,
            node_color=self.CODE.get_node_color(path, nodelist=graph.nodes), # type: ignore
            alpha=node_alpha,
        )
        nx.draw_networkx_edges(
            graph,
            pos=self.CODE.get_pos(nodelist=graph.nodes),
            node_size=node_size,
            width=[width_multiplier * weight + weightless_edge_width
                   for _, _, weight in graph.edges.data('weight')], # type: ignore
            edge_color=[constants.RED if set(edge)<=set(path)
                        else 'k' for edge in graph.edges], # type: ignore
            alpha=edge_alpha,
            **kwargs_for_networkx_draw,
        )


from unittest import mock

import pytest

from localuf import Surface
from localuf.type_aliases import Edge
from localuf.decoders import UF
from localuf._schemes import Global


@pytest.fixture
def global3F():
    """Global scheme for distance-3 surface code under code capacity noise model."""
    code = Surface(3, 'code capacity', scheme='global batch')
    return code.SCHEME


@pytest.mark.parametrize("d", range(3, 7, 2), ids=lambda x: f"d{x}")
@pytest.mark.parametrize("n", range(1, 4), ids=lambda x: f"n{x}")
def test_run(d: int, n: int):
    global_scheme = Surface(
        d,
        'phenomenological',
        scheme='global batch',
        window_height=d*n,
    ).SCHEME
    decoder = UF(global_scheme._CODE)
    noise_level = 0.5
    with (
        mock.patch('localuf._schemes.Global.reset') as mock_reset,
        mock.patch('localuf._schemes.Batch._sim_cycle_given_p') as scgp,
    ):
        assert global_scheme.run(decoder, noise_level, n) == (scgp.return_value, n)
        mock_reset.assert_called_once_with()
        scgp.assert_called_once_with(decoder, noise_level)


class TestGetLogicalError:

    def test_load_called(self, global3F: Global):
        e: Edge = ((0, 0), (0, 1))
        with mock.patch('localuf._pairs.Pairs.load') as mock_load:
            global3F.get_logical_error({e})
            mock_load.assert_called_once_with(e)

    def test_no_logical_error(self, global3F: Global):
        leftover: set[Edge] = {((0, j), (0, j+1)) for j in range(-1, 3-2)}
        assert global3F.get_logical_error(leftover) == 0

    def test_one_logical_error(self, global3F: Global):
        leftover: set[Edge] = {((0, j), (0, j+1)) for j in range(-1, 3-1)}
        assert global3F.get_logical_error(leftover) == 1

    def test_two_logical_errors(self, global3F: Global):
        leftover: set[Edge] = {((i, j), (i, j+1)) for j in range(-1, 3-1) for i in range(2)}
        assert global3F.get_logical_error(leftover) == 2


import pytest
from localuf._pairs import Pairs, LogicalCounter

@pytest.fixture
def repetition_counter():
    """Counter for distance-3 repetition code where node indices are (j, t)."""
    return LogicalCounter(
        d=3,
        commit_height=2,
        long_axis=0,
        time_axis=1,
    )

@pytest.fixture
def surface_counter():
    """Counter for distance-3 surface code where node indices are (i, j, t)."""
    return LogicalCounter(
        d=3,
        commit_height=2,
        long_axis=1,
        time_axis=2,
    )


def test_init():
    d = 5
    commit_height = 10
    long_axis = 1
    time_axis = 2
    logical_counter = LogicalCounter(d, commit_height, long_axis, time_axis)
    assert logical_counter._D == d
    assert logical_counter._COMMIT_HEIGHT == commit_height
    assert logical_counter._LONG_AXIS == long_axis
    assert logical_counter._TIME_AXIS == time_axis


class TestCount:

    def test_no_pairs(self, repetition_counter: LogicalCounter):
        pairs = Pairs()
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 0
        assert new_pairs._dc == {}

    def test_single_logical_error(self, repetition_counter: LogicalCounter):
        pairs = Pairs()
        pairs.add((-1, 0), (2, 0))
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 1
        assert new_pairs._dc == {}

    def test_multiple_logical_errors(self, repetition_counter: LogicalCounter):
        pairs = Pairs()
        pairs.add((-1, 0), (2, 0))
        pairs.add((-1, 1), (2, 1))
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 2
        assert new_pairs._dc == {}

    def test_no_logical_errors(self, repetition_counter: LogicalCounter):
        pairs = Pairs()
        pairs.add((0, 2), (1, 2))
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 0
        assert new_pairs.as_set == {((0, 0), (1, 0))}

    def test_mixed_pairs(self, repetition_counter: LogicalCounter):
        pairs = Pairs()
        pairs.add((-1, 0), (2, 0))
        pairs.add((-1, 1), (1, 2))
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 1
        assert new_pairs.as_set == {((-1, -1), (1, 0))}

    @pytest.mark.parametrize("j", (-1, 2))
    def test_ignore_pair_on_same_boundary(self, repetition_counter: LogicalCounter, j: int):
        pairs = Pairs()
        pairs.add((j, 0), (j, 1))
        error_count, new_pairs = repetition_counter.count(pairs)
        assert error_count == 0
        assert new_pairs._dc == {}

    def test_keep_zero_separation_pair(self, surface_counter: LogicalCounter):
        pairs = Pairs()
        pairs.add((0, 0, 2), (1, 0, 2))
        error_count, new_pairs = surface_counter.count(pairs)
        assert error_count == 0
        assert new_pairs.as_set == {((0, 0, 0), (1, 0, 0))}